//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <sstream>
#include <iomanip> // std::hex etc.
#include <iostream>

#include "nMaths.h"
#include "nCore.h"

static const char UnitName[] = "nCore";

//============================================================================
// Define global vars
//============================================================================

std::ostream& Erroutput = std::cerr;
int ioresult = 0;
int ExitCode = 0;

FFT_Array_type FFT_Array;

double FFT_raw_result[ MAX_FFT_LENGTH ];
double FFT_last_result[ MAX_FFT_LENGTH ];
double FFT_skew_result[ MAX_FFT_LENGTH ];
double OneOver[ MAX_FFT_LENGTH + 2 ];

BlockData_type BlockData;
timer_type timer;
Global_type Global;
parameters_type parameters;
settings_type settings;
process_type process;
results_type results;
history_type history;
LongDist_type LongDist;
SGNSFFT_type SGNSFFT;
strings_type strings;
version_type version;
PowersOf_type PowersOf;
Stats_type Stats;


void lossyWAVerror(std::string lwe_string, int lwe_value)
{
	if (! parameters.silent)
	{
		Write(Erroutput, "%lossyWAV Error% : ", lwe_string, CRLF);
	}

	throw (lwe_value);
}


void lossyWAVwarning(std::string lww_string)
{
	if ((parameters.warnings) && (! parameters.silent))
	{
		Write(Erroutput, "%lossyWAV Warning% : ", lww_string, CRLF);
	}
}


void GetVersionInfo()
{
// TODO // TY
/*
	std::string lptstrFilename;
	DWORD dwHandle;
	DWORD dwLen;
	DWORD puLen;
	void* lpData;
	void* lplpBuffer;
	strings.version = "";
	lptstrFilename = GetModuleName(hInstance);
	dwLen = GetFileVersionInfoSize(lptstrFilename.c_str(), dwHandle);

	if (dwLen > 0)
	{
		GetMem(lpData, dwLen);

		try
		{
			GetFileVersionInfo(lptstrFilename.c_str(), dwHandle, dwLen, lpData);
			VerQueryValue(lpData, '\\', lplpBuffer, puLen);
			version.MajorMinor = (* ((PVSFixedFileInfo) lplpBuffer)).dwFileVersionMS;
			version.ReleaseBuild = (* ((PVSFixedFileInfo) lplpBuffer)).dwFileVersionLS;
		}
		catch (...)
		{
			FreeMem(lpData);
		}

		strings.version = IntToStr(version.Major) + '.' +  IntToStr(version.Minor) + '.' +  IntToStr(version.Release) + '.' + IntToStr(version.Build);
	}
	else
	{
		version.MajorMinor = 0;
		version.ReleaseBuild = 0;
	}
*/
}


void gettimer()
{
	QueryPerformanceCounter(&timer.Stop);
	timer.Elapsed = (timer.Stop.QuadPart - timer.start.QuadPart) * timer.Period;

	if (timer.Elapsed != 0)
	{
		Global.processing_rate = Global.samples_processed * Global.sample_rate_recip / timer.Elapsed;
	}
}


void setpriority(int spv)
{
	SetPriorityClass(GetCurrentProcess, spv);
}


void ndelay()
{
	gettimer();
	timer.DelayEnd = int64_t(timer.Stop.QuadPart + timer.Frequency.QuadPart * timer.DelayTime);

	while (timer.Stop.QuadPart < timer.DelayEnd)
	{
		gettimer();
	}
}


std::string CardinalToHex(unsigned int cth_c)
{
	// output an 8 char uppercase hex string
	std::ostringstream ss;
	ss << std::hex << std::uppercase << std::setfill('0') << std::setw(8) << cth_c;
	return ss.str();
}


void write_unitname_and_datapointers(std::string UnitName, void* UserData1Pointer, void* UserData2Pointer, unsigned int UserData2Size)
{
	unsigned int datasize;
	datasize = ((unsigned int) UserData2Pointer) + UserData2Size - ((unsigned int) UserData1Pointer);

	Write(Erroutput, UnitName/*: 15*/, ':', CardinalToHex(((unsigned int) UserData1Pointer)), ',', CardinalToHex(((unsigned int) UserData2Pointer) + UserData2Size), ',', datasize/*: 10*/, " (", CardinalToHex(datasize), ") bytes.", CRLF);
}


void time_string_make(std::string& tsms, double tsmt)
{
	int tsmi, tsmk;

	if (tsmt < 3600)
	{
		tsmi = 4;
		tsmk = 100;
	}
	else
	{
		tsmi = 1;
		tsmk = 1;
	}

	tsmt = std::floor(tsmt * tsmk) * OneOver[tsmk] * Second_div_Day;

	datetimetostring(tsms, "hh:nn:ss.zzz", tsmt);
	tsms = tsms.substr(tsmi - 1, 8); // adjust for base 1
}


void size_string_make(std::string& ssms, double ssmt)
{
	const char size_chars[] = "kMGTPEZ";
	int ssmi, ssmj;
	std::string ssmx;
	double ssmy;

	if (ssmt > 99999)
	{
		ssmy = nlog2(ssmt) * OneOver[10];
		ssmj = int(ssmy);
		ssmi = int(log2_10 * (ssmy - ssmj));

		if (ssmi > 2)
		{
			ssmj ++;
			ssmi = 0;
		}

		ssmi = std::max(0, 3 - ssmi);
		ssmt = std::floor(ssmt * PowersOf.Two[- ssmj * 10] * PowersOf.Ten[ssmi]) * PowersOf.Ten[- ssmi];
		ssmx = size_chars[ssmj] + "iB";
	}
	else
	{
		ssmx = 'B';
		ssmi = 0;
	}

	ssms = floattostrf(ssmt, ffFixed, 99, ssmi) + ssmx;
}


void Set_FFT_Data_Rec(FFT_Data_Rec& FFTDataRec, int SFDR_Bits)
{
	/*with FFTDataRec do*/
	{
		FFTDataRec.bit_length = SFDR_Bits;
		FFTDataRec.bit_shift_from_max = MAX_FFT_BIT_LENGTH - FFTDataRec.bit_length;
		FFTDataRec.threshold_shift = (Global.Codec_Block_bits - FFTDataRec.bit_length) * SINGLE_HALF;
		FFTDataRec.length = PowersOf.TwoInt[FFTDataRec.bit_length];
		//FFTDataRec.length_recip = PowersOf.Two[- FFTDataRec.bit_length];   // repeated below! // TY
		FFTDataRec.length_m1 = FFTDataRec.length - 1;
		FFTDataRec.length_half = FFTDataRec.length >> 1;
		FFTDataRec.length_half_m1 = FFTDataRec.length_half - 1;
		FFTDataRec.length_recip = float(PowersOf.Two[- FFTDataRec.bit_length]);
		FFTDataRec.length_half_recip = FFTDataRec.length_recip * 2;
	}
}


void NSystem_Init()
{
	int nt_i, nt_j;

	for (nt_i = 0; nt_i <= 3; ++nt_i)
	{
		BlockData.Pointer[nt_i] = BlockData.Data[nt_i];
	}

	BlockData.Size.Prev = 0;
	BlockData.Size.Last = 0;
	BlockData.Size.This = 0;
	BlockData.Size.Next = 0;
	settings.scaling_factor = 1;
	settings.scaling_factor_inv = 1;
	GetVersionInfo();

	/*with version do*/
	if (version.Build < 27)
	{
		strings.version_short = IntToStr(version.Major) + '.' +  IntToStr(version.Minor) + '.' +  IntToStr(version.Release);

		if (version.Build > 0)
		{
			strings.version_short = strings.version_short + char(version.Build + 96);
		}
	}
	else
	{
		strings.version_short = strings.version;
	}

	timer.StartTime = std::time(NULL);
	QueryPerformanceFrequency(&timer.Frequency);
	QueryPerformanceCounter(&timer.start);
	timer.Period = 1.0 / timer.Frequency.QuadPart;

	for (nt_i = 0; nt_i < MAX_FFT_LENGTH; ++nt_i)
	{
		FFT_raw_result[nt_i] = 0;
	}

	for (nt_j = 0; nt_j < MAX_CHANNELS; nt_j ++)
	{
		LongDist.CHANNEL_OFFSET[nt_j] = nt_j * sizeof(LongDist.DATA_results[0]);
		history.CHANNEL_OFFSET[nt_j] = nt_j * sizeof(history.Last_Results[0]);
		SGNSFFT.CHANNEL_OFFSET[nt_j] = nt_j * sizeof(SGNSFFT.Last_Results[0]);

		for (nt_i = 0; nt_i <= MAX_HIST_FFT_SHORT_LENGTH / 2 + 1; ++nt_i)
		{
			history.DATA_results[nt_j][nt_i] = 0;
			history.BTRD_results[nt_j][nt_i] = 0;
			history.CORR_results[nt_j][nt_i] = 0;
			history.Last_Results[nt_j][nt_i] = 0;
		}

		for (nt_i = 0; nt_i <= MAX_FFT_LENGTH_HALF + 1; ++nt_i)
		{
			LongDist.DATA_results[nt_j][nt_i] = 0;
			LongDist.BTRD_results[nt_j][nt_i] = 0;
			LongDist.CORR_results[nt_j][nt_i] = 0;
		}

		for (nt_i = 0; nt_i <= 33; ++nt_i)
		{
			Stats.bits_removed[nt_j][nt_i] = 0;
		}

		for (nt_i = 1; nt_i <= PRECALC_ANALYSES + 2; ++nt_i)
		{
			results.minima[nt_j][nt_i] = 0;
		}
	}

	for (nt_i = 0; nt_i <= 1025; ++nt_i)
	{
		history.Histogram_DATA[nt_i] = 0;
		history.Histogram_BTRD[nt_i] = 0;
		history.Histogram_CORR[nt_i] = 0;
	}

	PowersOf.TwoInt[0] = 1;
	PowersOf.TwoM1[0] = 0;

	for (nt_i = 1; nt_i <= 62; ++nt_i)
	{
		PowersOf.TwoInt[nt_i] = PowersOf.TwoInt[nt_i - 1] + PowersOf.TwoInt[nt_i - 1];
		PowersOf.TwoM1[nt_i] = PowersOf.TwoInt[nt_i] - 1;
	}

	for (nt_i = -1024; nt_i <= 1023; ++nt_i)
	{
		PowersOf.Two[nt_i] = std::pow(2.0, nt_i);
	}

	for (nt_i = -308; nt_i <= 307; ++nt_i)
	{
		PowersOf.Ten[nt_i] = std::pow(10.0, nt_i);
	}

	OneOver[0] = 1;

	for (nt_i = 1; nt_i <= MAX_FFT_LENGTH + 1; ++nt_i)
	{
		OneOver[nt_i] = 1.0 / nt_i;
	}
}


static struct Init
{
	Init() {
		// initialization: todo
		NSystem_Init();
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
	}
} init;
