//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//
//     fourier.pas  -  Don Cross <dcross <at> intersrv <dot> com>
//
//     This is a Turbo Pascal Unit for calculating the n Fourier Transform
//     (FFT) and the Inverse n Fourier Transform (IFFT).
//     Visit the following URL for the latest version of this code.
//     This page also has a C/C++ version, and a brief discussion of the
//     theory behind the FFT algorithm.
//
//        http://www.intersrv.com/~dcross/fft.html#pascal [!!Dead link!!]
//
//     Revision history [most recent first]:
//
// 2012 Aug [Tyge Lvset (tycho), Aug. 2012]
//     Translated to C++ from Delphi
//
// 2009 May [Nick Currie]
//     Radix Eight added.
//
// 2009 April 24 [Nick Currie]
//     Radix Four added.
//
// 2008 September 29 [Nick Currie]
//     All code transcoded back to Delphi from IA-32/x87.
//
// 2008 May 12 [Nick Currie]
//     All code transcoded to IA-32/x87. FFT_Real implemented to allow calculation
//     of a packed Real array of length 2N to be calculated in a Complex array of
//     length N, with subsequent untangling. ReversedBits, Ai & Ar lookup tables
//     implemented. Calc_Frequency, FFT_Integer and FFT_Integer_Cleanup removed
//     as not used in lossyWAV.
//
// 1996 December 11 [Don Cross]
//     Improved documentation of the procedure CalcFrequency.
//     Fixed some messed up comments in procedure IFFT.
//
// 1996 December 6 [Don Cross]
//     Made procedure 'fft_integer' more efficient when buffer size changes
//     in successive calls:  the buffer is now only resized when the input
//     has more samples, not a differing number of samples.
//     Also changed the way 'fft_integer_cleanup' works so that it is
//     more "bullet-proof".
//
// 1996 December 4 [Don Cross]
//     Adding the procedure 'CalcFrequency', which calculates the FFT
//     at a specific frequency index p=0..n-1, instead of the whole
//     FFT.  This is O(n^2) instead of O(n*log(n)).
//
// 1996 November 30 [Don Cross]
//     Adding a routine to allow FFT of an input array of integers.
//     It is called 'fft_integer'.
//
// 1996 November 18 [Don Cross]
//     Added some comments.
//
// 1996 November 17 [Don Cross]
//     Wrote and debugged first version.
//
//==============================================================================

#include "nMaths.h"
#include "nFFT_pas.h"
#include "fftw_interface.h"
#include "nCore.h"

static const char UnitName [] = "nFFT_pas";

typedef tDComplex A_Rec [ MAX_FFT_LENGTH ];
typedef tDComplex* A_Rec_Ptr;
typedef unsigned char Bytes [ 4 ];
//typedef unsigned short Words [ 2 ];


unsigned char REVERSED_BYTES [ 256 ];
int NumSamples;
int NumberOfBitsNeeded;
double NumSamples_Reciprocal;
int OmegaStride;

#if defined( RB_LUT )
int R_Arr [ MAX_FFT_LENGTH ];
#else
unsigned short REVERSED_WORDS[ 65536 ];
#endif

tDComplex A_Arr [ MAX_FFT_LENGTH ];
int FullOmegaStride;

int SetFFTLength(int SFL_Samples)
{
	int result;

	if ((SFL_Samples & (SFL_Samples - 1)) == 0)
	{
		NumSamples = SFL_Samples;
		NumberOfBitsNeeded = int(nlog2(NumSamples));
		NumSamples_Reciprocal = double(PowersOf.Two[- NumberOfBitsNeeded]);
		result = NumSamples;
	}
	else
	{
		NumSamples = 0;
		NumberOfBitsNeeded = 0;
		NumSamples_Reciprocal = 0.0;
		result = 0;
	}

	FullOmegaStride = PowersOf.TwoInt[MAX_FFT_BIT_LENGTH - NumberOfBitsNeeded - 1];
	return result;
}


void shuffle_in_place_dcomplex()
{
	int sp_i, sp_j;
	tDComplex sp_c;
	OmegaStride = PowersOf.TwoInt[MAX_FFT_BIT_LENGTH - NumberOfBitsNeeded];

	for (sp_i = 1; sp_i <= NumSamples - 2; ++sp_i)
	{
#ifdef RB_LUT
		sp_j = R_Arr[sp_i] >> (MAX_FFT_BIT_LENGTH - NumberOfBitsNeeded);
#else
		sp_j = ((REVERSED_WORDS[((short*)&sp_i)[0]] << 16) | REVERSED_WORDS[((short*)&sp_i)[1]]) >> (32 - NumberOfBitsNeeded);
#endif

		if (sp_j < sp_i)
		{
			// std::swap(FFT_Array.DComplex[sp_i], FFT_Array.DComplex[sp_j]);
			sp_c = FFT_Array.DComplex[sp_i];
			FFT_Array.DComplex[sp_i] = FFT_Array.DComplex[sp_j];
			FFT_Array.DComplex[sp_j] = sp_c;
		}
	}
}


void RadixEight_DIT(int& BlockBitLen)
{
	int MainBlockSize;
	int DataStride;
	int D0, D1, D2, D3, D4, D5, D6, D7;
	int i, n, P, Q;
	tDComplex V0, V1, V2, V3, V4, V5, V6, V7;
	tDComplex X0, X1, X2, X3, X4, X5, X6, X7;
	tDComplex Y1, Y2, Y3, Y5, Y6, Y7;
	tDComplex U0, U1, U2, U3;
	DataStride = PowersOf.TwoInt[BlockBitLen];
	BlockBitLen = BlockBitLen + 3;
	OmegaStride = PowersOf.TwoInt[MAX_FFT_BIT_LENGTH - BlockBitLen];
	MainBlockSize = PowersOf.TwoInt[BlockBitLen];
	i = 0;

	do
	{
		D0 = i;
		D1 = D0 + DataStride;
		D2 = D1 + DataStride;
		D3 = D2 + DataStride;
		D4 = D3 + DataStride;
		D5 = D4 + DataStride;
		D6 = D5 + DataStride;
		D7 = D6 + DataStride;
		V0.Re = FFT_Array.DComplex[D0].Re;
		V0.Im = FFT_Array.DComplex[D0].Im;
		V1.Re = FFT_Array.DComplex[D4].Re;
		V1.Im = FFT_Array.DComplex[D4].Im;
		V2.Re = FFT_Array.DComplex[D2].Re;
		V2.Im = FFT_Array.DComplex[D2].Im;
		V3.Re = FFT_Array.DComplex[D6].Re;
		V3.Im = FFT_Array.DComplex[D6].Im;
		V4.Re = FFT_Array.DComplex[D1].Re;
		V4.Im = FFT_Array.DComplex[D1].Im;
		V5.Re = FFT_Array.DComplex[D5].Re;
		V5.Im = FFT_Array.DComplex[D5].Im;
		V6.Re = FFT_Array.DComplex[D3].Re;
		V6.Im = FFT_Array.DComplex[D3].Im;
		V7.Re = FFT_Array.DComplex[D7].Re;
		V7.Im = FFT_Array.DComplex[D7].Im;
		X0.Re = V0.Re + V4.Re;
		X0.Im = V0.Im + V4.Im;
		X1.Re = V1.Re + V7.Re;
		X1.Im = V1.Im + V7.Im;
		X2.Re = V2.Re + V6.Re;
		X2.Im = V2.Im + V6.Im;
		X3.Re = V3.Re + V5.Re;
		X3.Im = V3.Im + V5.Im;
		X4.Re = V0.Re - V4.Re;
		X4.Im = V0.Im - V4.Im;
		X5.Re = V1.Re - V7.Re;
		X5.Im = V1.Im - V7.Im;
		X6.Re = V2.Re - V6.Re;
		X6.Im = V2.Im - V6.Im;
		X7.Re = V3.Re - V5.Re;
		X7.Im = V3.Im - V5.Im;
		U0.Re = X0.Re + X2.Re;
		U0.Im = X0.Im + X2.Im;
		U1.Re = X1.Re + X3.Re;
		U1.Im = X1.Im + X3.Im;
		U2.Re = OneOverRoot2 * (X1.Re - X3.Re);
		U2.Im = OneOverRoot2 * (X1.Im - X3.Im);
		U3.Re = OneOverRoot2 * (X5.Re + X7.Re);
		U3.Im = OneOverRoot2 * (X5.Im + X7.Im);
		Y1.Re = X4.Re + U2.Re;
		Y1.Im = X4.Im + U2.Im;
		Y2.Re = X0.Re - X2.Re;
		Y2.Im = X0.Im - X2.Im;
		Y3.Re = X4.Re - U2.Re;
		Y3.Im = X4.Im - U2.Im;
		Y5.Re = U3.Re + X6.Re;
		Y5.Im = U3.Im + X6.Im;
		Y6.Re = X5.Re - X7.Re;
		Y6.Im = X5.Im - X7.Im;
		Y7.Re = U3.Re - X6.Re;
		Y7.Im = U3.Im - X6.Im;
		FFT_Array.DComplex[D0].Re = (U0.Re + U1.Re);
		FFT_Array.DComplex[D0].Im = (U0.Im + U1.Im);
		FFT_Array.DComplex[D1].Re = (Y1.Re + Y5.Im);
		FFT_Array.DComplex[D1].Im = (Y1.Im - Y5.Re);
		FFT_Array.DComplex[D2].Re = (Y2.Re + Y6.Im);
		FFT_Array.DComplex[D2].Im = (Y2.Im - Y6.Re);
		FFT_Array.DComplex[D3].Re = (Y3.Re + Y7.Im);
		FFT_Array.DComplex[D3].Im = (Y3.Im - Y7.Re);
		FFT_Array.DComplex[D4].Re = (U0.Re - U1.Re);
		FFT_Array.DComplex[D4].Im = (U0.Im - U1.Im);
		FFT_Array.DComplex[D5].Re = (Y3.Re - Y7.Im);
		FFT_Array.DComplex[D5].Im = (Y3.Im + Y7.Re);
		FFT_Array.DComplex[D6].Re = (Y2.Re - Y6.Im);
		FFT_Array.DComplex[D6].Im = (Y2.Im + Y6.Re);
		FFT_Array.DComplex[D7].Re = (Y1.Re - Y5.Im);
		FFT_Array.DComplex[D7].Im = (Y1.Im + Y5.Re);

		for (n = 1; n < DataStride; ++n)
		{
			D0 = i + n;
			D1 = D0 + DataStride;
			D2 = D1 + DataStride;
			D3 = D2 + DataStride;
			D4 = D3 + DataStride;
			D5 = D4 + DataStride;
			D6 = D5 + DataStride;
			D7 = D6 + DataStride;
			P = n * OmegaStride;
			Q = P;
			V0.Re = FFT_Array.DComplex[D0].Re;
			V0.Im = FFT_Array.DComplex[D0].Im;
			V1.Re = A_Arr[P].Re * FFT_Array.DComplex[D4].Re - A_Arr[P].Im * FFT_Array.DComplex[D4].Im;
			V1.Im = A_Arr[P].Re * FFT_Array.DComplex[D4].Im + A_Arr[P].Im * FFT_Array.DComplex[D4].Re;
			P = P + Q;
			V2.Re = A_Arr[P].Re * FFT_Array.DComplex[D2].Re - A_Arr[P].Im * FFT_Array.DComplex[D2].Im;
			V2.Im = A_Arr[P].Re * FFT_Array.DComplex[D2].Im + A_Arr[P].Im * FFT_Array.DComplex[D2].Re;
			P = P + Q;
			V3.Re = A_Arr[P].Re * FFT_Array.DComplex[D6].Re - A_Arr[P].Im * FFT_Array.DComplex[D6].Im;
			V3.Im = A_Arr[P].Re * FFT_Array.DComplex[D6].Im + A_Arr[P].Im * FFT_Array.DComplex[D6].Re;
			P = P + Q;
			V4.Re = A_Arr[P].Re * FFT_Array.DComplex[D1].Re - A_Arr[P].Im * FFT_Array.DComplex[D1].Im;
			V4.Im = A_Arr[P].Re * FFT_Array.DComplex[D1].Im + A_Arr[P].Im * FFT_Array.DComplex[D1].Re;
			P = P + Q;
			V5.Re = A_Arr[P].Re * FFT_Array.DComplex[D5].Re - A_Arr[P].Im * FFT_Array.DComplex[D5].Im;
			V5.Im = A_Arr[P].Re * FFT_Array.DComplex[D5].Im + A_Arr[P].Im * FFT_Array.DComplex[D5].Re;
			P = P + Q;
			V6.Re = A_Arr[P].Re * FFT_Array.DComplex[D3].Re - A_Arr[P].Im * FFT_Array.DComplex[D3].Im;
			V6.Im = A_Arr[P].Re * FFT_Array.DComplex[D3].Im + A_Arr[P].Im * FFT_Array.DComplex[D3].Re;
			P = P + Q;
			V7.Re = A_Arr[P].Re * FFT_Array.DComplex[D7].Re - A_Arr[P].Im * FFT_Array.DComplex[D7].Im;
			V7.Im = A_Arr[P].Re * FFT_Array.DComplex[D7].Im + A_Arr[P].Im * FFT_Array.DComplex[D7].Re;
			X0.Re = V0.Re + V4.Re;
			X0.Im = V0.Im + V4.Im;
			X1.Re = V1.Re + V7.Re;
			X1.Im = V1.Im + V7.Im;
			X2.Re = V2.Re + V6.Re;
			X2.Im = V2.Im + V6.Im;
			X3.Re = V3.Re + V5.Re;
			X3.Im = V3.Im + V5.Im;
			X4.Re = V0.Re - V4.Re;
			X4.Im = V0.Im - V4.Im;
			X5.Re = V1.Re - V7.Re;
			X5.Im = V1.Im - V7.Im;
			X6.Re = V2.Re - V6.Re;
			X6.Im = V2.Im - V6.Im;
			X7.Re = V3.Re - V5.Re;
			X7.Im = V3.Im - V5.Im;
			U0.Re = X0.Re + X2.Re;
			U0.Im = X0.Im + X2.Im;
			U1.Re = X1.Re + X3.Re;
			U1.Im = X1.Im + X3.Im;
			U2.Re = OneOverRoot2 * (X1.Re - X3.Re);
			U2.Im = OneOverRoot2 * (X1.Im - X3.Im);
			U3.Re = OneOverRoot2 * (X5.Re + X7.Re);
			U3.Im = OneOverRoot2 * (X5.Im + X7.Im);
			Y1.Re = X4.Re + U2.Re;
			Y1.Im = X4.Im + U2.Im;
			Y2.Re = X0.Re - X2.Re;
			Y2.Im = X0.Im - X2.Im;
			Y3.Re = X4.Re - U2.Re;
			Y3.Im = X4.Im - U2.Im;
			Y5.Re = U3.Re + X6.Re;
			Y5.Im = U3.Im + X6.Im;
			Y6.Re = X5.Re - X7.Re;
			Y6.Im = X5.Im - X7.Im;
			Y7.Re = U3.Re - X6.Re;
			Y7.Im = U3.Im - X6.Im;
			FFT_Array.DComplex[D0].Re = (U0.Re + U1.Re);
			FFT_Array.DComplex[D0].Im = (U0.Im + U1.Im);
			FFT_Array.DComplex[D1].Re = (Y1.Re + Y5.Im);
			FFT_Array.DComplex[D1].Im = (Y1.Im - Y5.Re);
			FFT_Array.DComplex[D2].Re = (Y2.Re + Y6.Im);
			FFT_Array.DComplex[D2].Im = (Y2.Im - Y6.Re);
			FFT_Array.DComplex[D3].Re = (Y3.Re + Y7.Im);
			FFT_Array.DComplex[D3].Im = (Y3.Im - Y7.Re);
			FFT_Array.DComplex[D4].Re = (U0.Re - U1.Re);
			FFT_Array.DComplex[D4].Im = (U0.Im - U1.Im);
			FFT_Array.DComplex[D5].Re = (Y3.Re - Y7.Im);
			FFT_Array.DComplex[D5].Im = (Y3.Im + Y7.Re);
			FFT_Array.DComplex[D6].Re = (Y2.Re - Y6.Im);
			FFT_Array.DComplex[D6].Im = (Y2.Im + Y6.Re);
			FFT_Array.DComplex[D7].Re = (Y1.Re - Y5.Im);
			FFT_Array.DComplex[D7].Im = (Y1.Im + Y5.Re);
		}

		i = i + MainBlockSize;
	}
	while (!(i == NumSamples));
}

void RadixFour_DIT(int& BlockBitLen)
{
	int MainBlockSize;
	int DataStride;
	int D0, D1, D2, D3;
	int i, n;
	tDComplex V0, V1, V2, V3;
	tDComplex X0, X1, X2, X3;
	DataStride = PowersOf.TwoInt[BlockBitLen];
	BlockBitLen = BlockBitLen + 2;
	OmegaStride = PowersOf.TwoInt[MAX_FFT_BIT_LENGTH - BlockBitLen];
	MainBlockSize = PowersOf.TwoInt[BlockBitLen];
	i = 0;

	do
	{
		D0 = i;
		D1 = D0 + DataStride;
		D2 = D1 + DataStride;
		D3 = D2 + DataStride;
		V0.Re = FFT_Array.DComplex[D0].Re;
		V0.Im = FFT_Array.DComplex[D0].Im;
		V1.Re = FFT_Array.DComplex[D2].Re;
		V1.Im = FFT_Array.DComplex[D2].Im;
		V2.Re = FFT_Array.DComplex[D1].Re;
		V2.Im = FFT_Array.DComplex[D1].Im;
		V3.Re = FFT_Array.DComplex[D3].Re;
		V3.Im = FFT_Array.DComplex[D3].Im;
		X0.Re = V0.Re + V2.Re;
		X0.Im = V0.Im + V2.Im;
		X1.Re = V1.Re + V3.Re;
		X1.Im = V1.Im + V3.Im;
		X2.Re = V0.Re - V2.Re;
		X2.Im = V0.Im - V2.Im;
		X3.Re = V3.Im - V1.Im;
		X3.Im = V1.Re - V3.Re;
		FFT_Array.DComplex[D0].Re = (X0.Re + X1.Re);
		FFT_Array.DComplex[D0].Im = (X0.Im + X1.Im);
		FFT_Array.DComplex[D1].Re = (X2.Re - X3.Re);
		FFT_Array.DComplex[D1].Im = (X2.Im - X3.Im);
		FFT_Array.DComplex[D2].Re = (X0.Re - X1.Re);
		FFT_Array.DComplex[D2].Im = (X0.Im - X1.Im);
		FFT_Array.DComplex[D3].Re = (X2.Re + X3.Re);
		FFT_Array.DComplex[D3].Im = (X2.Im + X3.Im);

		for (n = 1; n < DataStride; ++n)
		{
			D0 = i + n;
			D1 = D0 + DataStride;
			D2 = D1 + DataStride;
			D3 = D2 + DataStride;
			V0.Re = FFT_Array.DComplex[D0].Re;
			V0.Im = FFT_Array.DComplex[D0].Im;
			V1.Re = A_Arr[n * OmegaStride].Re * FFT_Array.DComplex[D2].Re - A_Arr[n * OmegaStride].Im * FFT_Array.DComplex[D2].Im;
			V1.Im = A_Arr[n * OmegaStride].Re * FFT_Array.DComplex[D2].Im + A_Arr[n * OmegaStride].Im * FFT_Array.DComplex[D2].Re;
			V2.Re = A_Arr[n * 2 * OmegaStride].Re * FFT_Array.DComplex[D1].Re - A_Arr[n * 2 * OmegaStride].Im * FFT_Array.DComplex[D1].Im;
			V2.Im = A_Arr[n * 2 * OmegaStride].Re * FFT_Array.DComplex[D1].Im + A_Arr[n * 2 * OmegaStride].Im * FFT_Array.DComplex[D1].Re;
			V3.Re = A_Arr[n * 3 * OmegaStride].Re * FFT_Array.DComplex[D3].Re - A_Arr[n * 3 * OmegaStride].Im * FFT_Array.DComplex[D3].Im;
			V3.Im = A_Arr[n * 3 * OmegaStride].Re * FFT_Array.DComplex[D3].Im + A_Arr[n * 3 * OmegaStride].Im * FFT_Array.DComplex[D3].Re;
			X0.Re = V0.Re + V2.Re;
			X0.Im = V0.Im + V2.Im;
			X1.Re = V1.Re + V3.Re;
			X1.Im = V1.Im + V3.Im;
			X2.Re = V0.Re - V2.Re;
			X2.Im = V0.Im - V2.Im;
			X3.Re = V3.Im - V1.Im;
			X3.Im = V1.Re - V3.Re;
			FFT_Array.DComplex[D0].Re = (X0.Re + X1.Re);
			FFT_Array.DComplex[D0].Im = (X0.Im + X1.Im);
			FFT_Array.DComplex[D1].Re = (X2.Re - X3.Re);
			FFT_Array.DComplex[D1].Im = (X2.Im - X3.Im);
			FFT_Array.DComplex[D2].Re = (X0.Re - X1.Re);
			FFT_Array.DComplex[D2].Im = (X0.Im - X1.Im);
			FFT_Array.DComplex[D3].Re = (X2.Re + X3.Re);
			FFT_Array.DComplex[D3].Im = (X2.Im + X3.Im);
		}

		i = i + MainBlockSize;
	}
	while (!(i == NumSamples));
}

void RadixTwo_DIT(int& BlockBitLen)
{
	int MainBlockSize;
	int DataStride;
	int D0, D1;
	int i, n;
	tDComplex V0, V1;
	DataStride = PowersOf.TwoInt[BlockBitLen];
	BlockBitLen = BlockBitLen + 1;
	OmegaStride = PowersOf.TwoInt[MAX_FFT_BIT_LENGTH - BlockBitLen];
	MainBlockSize = PowersOf.TwoInt[BlockBitLen];
	i = 0;

	do
	{
		D0 = i;
		D1 = D0 + DataStride;
		V0.Re = FFT_Array.DComplex[D0].Re;
		V0.Im = FFT_Array.DComplex[D0].Im;
		V1.Re = FFT_Array.DComplex[D1].Re;
		V1.Im = FFT_Array.DComplex[D1].Im;
		FFT_Array.DComplex[D0].Re = V0.Re + V1.Re;
		FFT_Array.DComplex[D0].Im = V0.Im + V1.Im;
		FFT_Array.DComplex[D1].Re = V0.Re - V1.Re;
		FFT_Array.DComplex[D1].Im = V0.Im - V1.Im;

		for (n = 1; n <= DataStride; ++n)
		{
			D0 = i + n;
			D1 = D0 + DataStride;
			V0.Re = FFT_Array.DComplex[D0].Re;
			V0.Im = FFT_Array.DComplex[D0].Im;
			V1.Re = A_Arr[n * OmegaStride].Re * FFT_Array.DComplex[D1].Re - A_Arr[n * OmegaStride].Im * FFT_Array.DComplex[D1].Im;
			V1.Im = A_Arr[n * OmegaStride].Re * FFT_Array.DComplex[D1].Im + A_Arr[n * OmegaStride].Im * FFT_Array.DComplex[D1].Re;
			FFT_Array.DComplex[D0].Re = V0.Re + V1.Re;
			FFT_Array.DComplex[D0].Im = V0.Im + V1.Im;
			FFT_Array.DComplex[D1].Re = V0.Re - V1.Re;
			FFT_Array.DComplex[D1].Im = V0.Im - V1.Im;
		}

		i = i + MainBlockSize;
	}
	while (!(i == NumSamples));
}

void FFT_DIT_Complex()
{
	int BlockBitLen;

	shuffle_in_place_dcomplex();
	BlockBitLen = 0;

	while (NumberOfBitsNeeded - BlockBitLen > 2)
	{
		RadixEight_DIT(BlockBitLen);
	}

	while (NumberOfBitsNeeded - BlockBitLen > 1)
	{
		RadixFour_DIT(BlockBitLen);
	}

	while (NumberOfBitsNeeded - BlockBitLen > 0)
	{
		RadixTwo_DIT(BlockBitLen);
	}
}


void FFT_DIT_Real()
{
	tDComplex V1, X0, X1;
	int J, k, l, m, n;
	FFT_DIT_Complex();
	J = 0;
	n = PowersOf.TwoInt[NumberOfBitsNeeded - 1];

	do
	{
		++J;
		k = NumSamples - J;
		l = NumSamples + k;
		m = NumSamples + J;
		X0.Re = SINGLE_HALF * (FFT_Array.DComplex[J].Re + FFT_Array.DComplex[k].Re);
		X0.Im = SINGLE_HALF * (FFT_Array.DComplex[J].Im + FFT_Array.DComplex[k].Im);
		X1.Re = SINGLE_HALF * (FFT_Array.DComplex[J].Re - FFT_Array.DComplex[k].Re);
		X1.Im = SINGLE_HALF * (FFT_Array.DComplex[J].Im - FFT_Array.DComplex[k].Im);
		V1.Re = A_Arr[J * FullOmegaStride].Re * X0.Im + A_Arr[J * FullOmegaStride].Im * X1.Re;
		V1.Im = A_Arr[J * FullOmegaStride].Re * X1.Re - A_Arr[J * FullOmegaStride].Im * X0.Im;
		FFT_Array.DComplex[J].Re = (X0.Re + V1.Re);
		FFT_Array.DComplex[J].Im = (X1.Im - V1.Im);
		FFT_Array.DComplex[l].Re = FFT_Array.DComplex[J].Re;
		FFT_Array.DComplex[l].Im = -FFT_Array.DComplex[J].Im;
		FFT_Array.DComplex[m].Re = (X0.Re - V1.Re);
		FFT_Array.DComplex[m].Im = (X1.Im + V1.Im);
		FFT_Array.DComplex[k].Re = FFT_Array.DComplex[m].Re;
		FFT_Array.DComplex[k].Im = -FFT_Array.DComplex[m].Im;
	}
	while (!(J == n));

	FFT_Array.DComplex[NumSamples].Re = FFT_Array.DComplex[0].Re - FFT_Array.DComplex[0].Im;
	FFT_Array.DComplex[NumSamples].Im = 0;
	FFT_Array.DComplex[0].Re = FFT_Array.DComplex[0].Re + FFT_Array.DComplex[0].Im;
	FFT_Array.DComplex[0].Im = 0;
}


void IFFT_DIT_Complex()
{
	int BlockBitLen;
	int MainBlockSize;
	int DataStride;
	int i, J, k, n;
	tDComplex V;
	shuffle_in_place_dcomplex();

	for (BlockBitLen = 1; BlockBitLen <= NumberOfBitsNeeded; ++BlockBitLen)
	{
		i = 0;
		OmegaStride = PowersOf.TwoInt[MAX_FFT_BIT_LENGTH - BlockBitLen];
		MainBlockSize = PowersOf.TwoInt[BlockBitLen];
		DataStride = MainBlockSize >> 1;

		do
		{
			J = i;

			for (n = 0; n < DataStride; ++n)
			{
				k = J + DataStride;
	
				tDComplex& Cn = A_Arr[n * OmegaStride];
				tDComplex& Ck = FFT_Array.DComplex[k];
				tDComplex& CJ = FFT_Array.DComplex[J];
				V.Re = (Cn.Re * Ck.Re) + (Cn.Im * Ck.Im);
				V.Im = (Cn.Re * Ck.Im) - (Cn.Im * Ck.Re);
				Ck.Re = CJ.Re - V.Re;
				Ck.Im = CJ.Im - V.Im;
				CJ.Re += V.Re;
				CJ.Im += V.Im;
/*
				V.Re = A_Arr[n * OmegaStride].Re * FFT_Array.DComplex[k].Re + A_Arr[n * OmegaStride].Im * FFT_Array.DComplex[k].Im;
				V.Im = A_Arr[n * OmegaStride].Re * FFT_Array.DComplex[k].Im - A_Arr[n * OmegaStride].Im * FFT_Array.DComplex[k].Re;
				FFT_Array.DComplex[k].Re = FFT_Array.DComplex[J].Re - V.Re;
				FFT_Array.DComplex[k].Im = FFT_Array.DComplex[J].Im - V.Im;
				FFT_Array.DComplex[J].Re = FFT_Array.DComplex[J].Re + V.Re;
				FFT_Array.DComplex[J].Im = FFT_Array.DComplex[J].Im + V.Im;
*/
				++J;
			}

			i += MainBlockSize;
		}
		while (!(i == NumSamples));
	}
}


void nfft_double_init()
{
	int nf_i, nf_j;
	double delta;

	//for nf_i:=0 to MAX_FFT_LENGTH*2-1 do
	//  FFT_Array.Real[nf_i]:=0;

	//==========================================================================
	// Create 256 byte reversed-bits lookup table
	//==========================================================================
	for (nf_i = 0; nf_i <= 15; ++nf_i)
	{
		for (nf_j = 0; nf_j <= 15; ++nf_j)
		{
			REVERSED_BYTES[(nf_j << 4) + nf_i] = REVERSED_NIBBLES[nf_j] | (REVERSED_NIBBLES[nf_i] << 4);
		}
	}

#if ! defined( RB_LUT )

	for (nf_i = 0; nf_i <= 65535; ++nf_i)
	{
		REVERSED_WORDS[nf_i] = (REVERSED_BYTES[nf_i & 255] << 8) | REVERSED_BYTES[nf_i >> 8];
	}

#endif
	delta = (-TwoPi) * PowersOf.Two[- MAX_FFT_BIT_LENGTH];

	for (nf_i = 0; nf_i < MAX_FFT_LENGTH; ++nf_i)
	{
#if defined( RB_LUT )
		R_Arr[nf_i] = ((((REVERSED_BYTES[((Bytes) nf_i)[0]] << 8) + (REVERSED_BYTES[((Bytes) nf_i)[1]]) << 8) + (REVERSED_BYTES[((Bytes) nf_i)[2]]) << 8) + REVERSED_BYTES[((Bytes) nf_i)[3]]) >> (32 - MAX_FFT_BIT_LENGTH);
#endif
		/*with A_Arr[nf_i] do*/
		{	
			A_Arr[nf_i].Re = std::cos(nf_i * delta);
			A_Arr[nf_i].Im = std::sin(nf_i * delta);
		}
	}
}


static struct Init
{
	Init() {
		// initialization: todo
		if (! FFTW_Initialised())
		{
			nfft_double_init();
		}
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
	}
} init;
