//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include "nCore.h"
#include "nSpreading.h"
#include "nParameter.h"
#include "nInitialise.h"
#include "nSGNS.h"
#include "nProcess.h"
#include "nFillFFT.h"
#include "nRemoveBits.h"

static const char UnitName [] = "nInitialise";

//============================================================================
// Quality Preset Data.
//============================================================================
static const int QUALITY_PRESET_MIN = -5;
static const int QUALITY_PRESET_MAX = 10;

//============================================================================
// constants directly related to quality_integer settings.
//============================================================================                   //      -5,      -4,      -3,      -2,      -1,       0,       1,       2,       3,       4,       5,       6,       7,       8,       9,      10
const range_array<float, QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_MINIMUM_BITS_TO_KEEP   = {   2.710f,   2.848f,   2.986f,   3.124f,   3.262f,   3.400f,   3.470f,   3.540f,   3.610f,   3.680f,   3.750f,   4.000f,   4.250f,   4.500f,   4.750f,   5.000f };
const range_array<float, QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_NOISE_THRESHOLD_SHIFTS = {  16.667f,  14.444f,  12.222f,  10.000f,   7.778f,   5.556f,   4.444f,   3.333f,   2.222f,   1.111f,   0.000f,  -2.222f,  -4.444f,  -6.667f,  -8.889f, -11.111f };
const range_array<float, QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_SIGNAL_TO_NOISE_RATIOS = { -19.500f, -20.500f, -21.500f, -22.500f, -23.500f, -24.500f, -25.000f, -25.500f, -26.000f, -26.500f, -27.000f, -28.000f, -29.000f, -30.000f, -31.000f, -32.000f };
const range_array<int,   QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_CLIPS_PER_CHANNEL      = {        3,        3,        3,        3,        3,        2,        2,        2,        2,        2,        1,        1,        1,        0,        0,        0 };
const range_array<int,   QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_UPPER_CALC_FREQ_LIMIT  = {    15159,    15332,    15504,    15676,    15848,    16021,    16107,    16193,    16279,    16365,    16451,    16538,    16624,    16710,    16796,    16882 };
const range_array<float, QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_AUTO_SHAPING_FACTOR    = {0.077903f,0.104718f,0.133335f,0.162936f,0.195383f,0.239690f,0.265891f,0.304356f,0.340499f,0.380561f,0.422840f,0.517277f,0.621117f,0.734487f,0.863206f,1.000000f };

//============================================================================
// Quality_Noise_Threshold_Delta is the correction to the calculated noise
// average when using DC offset corrected FFT input data (default).
//============================================================================
const range_array<float, QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_NOISE_THRESHOLD_DELTA  = {2.420359f,2.296165f,2.141537f,1.964923f,1.753599f,1.424952f,1.337800f,1.191097f,1.096227f,0.923858f,0.765617f,0.350196f,0.000000f,0.000000f,0.000000f,0.000000 };
//const range_array<float, QUALITY_PRESET_MIN, QUALITY_PRESET_MAX> QUALITY_NOISE_THRESHOLD_DELTA  = {  2.4240f,  2.3010f,  2.1370f,  1.9580f,  1.7710f,  1.4130f,  1.3200f,  1.1850f,  1.0970f,  0.9290f,  0.7740f,  0.3780f,  0.0000f,  0.0000f,  0.0000f,  0.0000 };


//============================================================================
// Check command line parameters from nParameter;
//============================================================================
void Check_Switches()
{
	settings.quality_double = parameters.quality;
	settings.quality_integer = int(settings.quality_double);
	settings.quality_fraction = settings.quality_double - settings.quality_integer;

	if (parameters.analyses != 3)
	{
		strings.parameter += " --analyses ";
		strings.parameter += IntToStr(parameters.analyses);
	}

	if (parameters.scaling != -1)
	{
		settings.scaling_factor = parameters.scaling;
		strings.parameter += " --scale ";
		strings.parameter += floattostrf(settings.scaling_factor, ffFixed, 99, 6);
	}
	else
	{
		settings.scaling_factor = 1.0;
	}

	settings.scaling_factor_inv = 1.0 / settings.scaling_factor;

	if (parameters.linkchannels || parameters.midside)
	{
		strings.parameter += " --linkchannels";
	}

	if (parameters.shaping > -1)
	{
		strings.parameter += " --shaping ";
		strings.parameter += floattostrf(parameters.shaping, ffFixed, 99, 4);
	}

	if ((parameters.adaptive_shaping == -1) || (parameters.adaptive_shaping == -3))
	{
		strings.parameter += " --adaptive off";
	}
	else if ((parameters.adaptive_shaping != -2) || (! parameters.adaptive_shaping_warp))
	{
		strings.parameter += " --adaptive";

		if (parameters.adaptive_shaping > 0)
		{
			strings.parameter += " ";
			strings.parameter += IntToStr(parameters.adaptive_shaping);
		}

		if (! parameters.adaptive_shaping_warp)
		{
			strings.parameter += " nowarp";
		}
	}

	if (parameters.midside)
	{
		strings.parameter += " --midside";
	}

	if (parameters.dccorrect)
	{
		settings.dccorrect_multiplier = 1;
	}
	else
	{
		settings.dccorrect_multiplier = 0;
		strings.parameter += " --nodccorrect";
	}

	if (parameters.underlap > 0)
	{
		strings.parameter += " --underlap ";
		strings.parameter += IntToStr(parameters.underlap);
	}

	if (parameters.maxclips > -1)
	{
		settings.maximum_clips_per_channel = parameters.maxclips;
		strings.parameter += " --maxclips ";
		strings.parameter += IntToStr(parameters.maxclips);
	}
	else
	{
		settings.maximum_clips_per_channel = QUALITY_CLIPS_PER_CHANNEL[std::min(QUALITY_PRESET_MAX, settings.quality_integer + int (settings.quality_fraction != 0))];
	}
}


//============================================================================
// Setup most of the internal arrays of values to be used in the analyses;
//============================================================================
void initial_setup()
{
	int sa_i;
	int this_channel;
	int this_analysis_number;
	double this_threshold_shift;
	double this_dccorrect_shift;
	double this_shaping_factor;
	settings.static_minimum_bits_to_keep = std::min(Global.bits_per_sample - 4, parameters.Static);

	if (parameters.Static > STATIC_MINIMUM_BITS_TO_KEEP)
	{
		strings.parameter += "--static ";
		strings.parameter += IntToStr(parameters.Static);
	}

	if (parameters.limit != -1)
	{
		Global.upper_freq_limit = std::min(int(Global.sample_rate * 0.453515), parameters.limit);
		strings.parameter += " --limit ";
		strings.parameter += floattostrf(Global.upper_freq_limit, ffFixed, 99, 0);
	}
	else
	{
		Global.upper_freq_limit = std::min(int(Global.sample_rate * 0.453515), QUALITY_UPPER_CALC_FREQ_LIMIT[settings.quality_integer]);
	}

	Frequency_Limits[SPREAD_ZONES + 1] = Global.upper_freq_limit;
	Global.lower_freq_limit = LOWER_FREQ_LIMIT;

	if (Global.channels <= 2)
	{
		parameters.perchannel = true;
	}

	history.Histogram_Multiplier = PowersOf.Two[6 - Global.bits_per_sample];
	history.Histogram_Length = 64;
	history.Histogram_Offset = history.Histogram_Length >> 1;

	//==========================================================================
	// Make an sample-rate adjustment to noise_threshold_shift.
	//==========================================================================
	this_threshold_shift = -db_lg2x20 * Global.Codec_Block_bit_shift * SINGLE_HALF;

	//==========================================================================
	// Set noise-thresholds, adjusting  for samplerate - NB: most tuning was
	// performed using 44.1kHz samples!! (second tweak)
	//==========================================================================
	if (settings.quality_integer == QUALITY_PRESET_MAX)
	{
		this_dccorrect_shift = QUALITY_NOISE_THRESHOLD_DELTA[QUALITY_PRESET_MAX];
		settings.noise_threshold_shift_minimum = QUALITY_NOISE_THRESHOLD_SHIFTS[QUALITY_PRESET_MAX];
		settings.noise_threshold_shift_average = QUALITY_SIGNAL_TO_NOISE_RATIOS[QUALITY_PRESET_MAX];
		settings.dynamic_minimum_bits_to_keep = QUALITY_MINIMUM_BITS_TO_KEEP[QUALITY_PRESET_MAX];
		this_shaping_factor = QUALITY_AUTO_SHAPING_FACTOR[QUALITY_PRESET_MAX];
	}
	else
	{
		this_dccorrect_shift = (settings.quality_fraction * (QUALITY_NOISE_THRESHOLD_DELTA[settings.quality_integer + 1] - QUALITY_NOISE_THRESHOLD_DELTA[settings.quality_integer]) + QUALITY_NOISE_THRESHOLD_DELTA[settings.quality_integer]);
		settings.noise_threshold_shift_minimum = settings.quality_fraction * (QUALITY_NOISE_THRESHOLD_SHIFTS[settings.quality_integer + 1] - QUALITY_NOISE_THRESHOLD_SHIFTS[settings.quality_integer]) + QUALITY_NOISE_THRESHOLD_SHIFTS[settings.quality_integer];
		settings.noise_threshold_shift_average = settings.quality_fraction * (QUALITY_SIGNAL_TO_NOISE_RATIOS[settings.quality_integer + 1] - QUALITY_SIGNAL_TO_NOISE_RATIOS[settings.quality_integer]) + QUALITY_SIGNAL_TO_NOISE_RATIOS[settings.quality_integer];
		settings.dynamic_minimum_bits_to_keep = settings.quality_fraction * (QUALITY_MINIMUM_BITS_TO_KEEP[settings.quality_integer + 1] - QUALITY_MINIMUM_BITS_TO_KEEP[settings.quality_integer]) + QUALITY_MINIMUM_BITS_TO_KEEP[settings.quality_integer];
		this_shaping_factor = settings.quality_fraction * (QUALITY_AUTO_SHAPING_FACTOR[settings.quality_integer + 1] - QUALITY_AUTO_SHAPING_FACTOR[settings.quality_integer]) + QUALITY_AUTO_SHAPING_FACTOR[settings.quality_integer];
	}

	this_dccorrect_shift *= settings.dccorrect_multiplier;
	settings.noise_threshold_shift_average += this_dccorrect_shift + this_threshold_shift;
	settings.noise_threshold_shift_minimum += this_threshold_shift;

	if (parameters.dynamic != -1)
	{
		strings.parameter += " --dynamic ";
		strings.parameter += floattostrf(parameters.dynamic, ffFixed, 99, 3);
		settings.dynamic_minimum_bits_to_keep = parameters.dynamic;
	}

	if (parameters.shaping == -2)
	{
		parameters.shaping = this_shaping_factor;
	}

	//==========================================================================
	// Select which FFT analyses to perform.
	//==========================================================================
	settings.FFT_analysis_switches = (SHORT_FFT_LENGTH | LONG_FFT_LENGTH);

	if (parameters.analyses > 2)
	{
		settings.FFT_analysis_switches |= IMPULSE_FFT_LENGTH;
	}

	if (parameters.analyses > 3)
	{
		settings.FFT_analysis_switches |= ADDED_FFT_LENGTH_512;
	}

	if (parameters.analyses > 4)
	{
		settings.FFT_analysis_switches |= ADDED_FFT_LENGTH_256;
	}

	if (parameters.analyses > 5)
	{
		settings.FFT_analysis_switches |= ADDED_FFT_LENGTH_128;
	}

	//==========================================================================
	// Shift analysis flags to reflect actual codec-block-size.
	//==========================================================================
	settings.FFT_analysis_switches = int(settings.FFT_analysis_switches * PowersOf.Two[Global.Codec_Block_bit_shift]);

	//==========================================================================
	Set_FFT_Data_Rec(history.FFT, SHORT_FFT_BIT_LENGTH + Global.Codec_Block_bit_shift);
	Set_FFT_Data_Rec(LongDist.FFT, Global.Codec_Block_bits);

	for (sa_i = 1; sa_i <= PRECALC_ANALYSES; ++sa_i)
	{
		Global.analysis.bits[sa_i] = history.FFT.bit_length - 2 + sa_i;
		Global.analysis.length[sa_i] = PowersOf.TwoInt[Global.analysis.bits[sa_i]];
	}

	if (parameters.freqdist || (parameters.adaptive_shaping != -1))
	{
		settings.raw_result_short = history.FFT.length;
	}
	else
	{
		settings.raw_result_short = 0;
	}

	if (parameters.adaptive_shaping != -1)
	{
		Set_FFT_Data_Rec(SGNSFFT.FFT, LONG_FFT_BIT_LENGTH + Global.Codec_Block_bit_shift);
		settings.raw_result_sgns = SGNSFFT.FFT.length;
		nSGNS_Initialise();
	}
	else
	{
		settings.raw_result_sgns = 0;
	}

	if (Global.upper_freq_limit > Global.sample_rate * 0.475)
	{
		Global.upper_freq_limit = int(Global.sample_rate * 0.475);
	}

	if (! parameters.silent)
	{
		if (parameters.verbosity)
		{
			Write(Erroutput, "Filename  : ", WAVFilePrintName(), CRLF, "File Info : ", Global.sample_rate * OneOver[1000]/*: 0: 2*/, "kHz; ", Global.channels, " channel; ", Global.bits_per_sample, " bit; ");

			if (Global.wave_size < MAX_WAVE_SIZE)
			{
				time_string_make(strings.time, double(Global.total_samples) / Global.sample_rate);
				size_string_make(strings.Size, double(Global.total_samples) * Global.channels * Global.bytes_per_sample);

				Write(Erroutput, strings.time, ", ", strings.Size);
			}

			Write(Erroutput, "\nProgress  : ");
		}
		else
		{
			Write(Erroutput, WAVFilePrintName(), ';');
		}
	}

	flush(Erroutput);

	for (this_channel = 0; this_channel < Global.channels; ++this_channel)
	{
		Global.channel = this_channel;

		for (this_analysis_number = 1; this_analysis_number <= PRECALC_ANALYSES; ++this_analysis_number)
		{
			results.saved_FFT_results[Global.channel][this_analysis_number].start = -1;
		}
	}

	Global.clipped_samples = 0;

	nProcess_Init();
	nSpreading_Init();
	nFillFFT_Init();	  // dependent on codec_block_size.
	nRemoveBits_Init();  // bitdepth and samplerate dependent.

	Global.output_blocks = int(PowersOf.Two[20 - Global.Codec_Block_bits] * OneOver[Global.channels]);
	Global.last_print = 0;

	Global.blocks_processed = 1;
	Global.blocks_processed_recip = 1;

	Global.bits_removed = 0;
	Global.bits_lost = 0;
}
// initialization: todo
// finalization: todo Write ( Erroutput, UnitName, CRLF );
