//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <cstdlib>

#include "nCore.h"
#include "nParameter.h"
#include "nMaths.h"

extern int    main_argc;
extern char** main_argv;


namespace {

const char UnitName [] = "nParameter";

const char lossyWAV_GPL [] =
    "This program is free software: you can redistribute it and/or modify it under\n"
    "the terms of the GNU General Public License as published by the Free Software\n"
    "Foundation, either version 3 of the License, or (at your option) any later\n"
    "version.\n\n"
    "This program is distributed in the hope that it will be useful,but WITHOUT ANY\n"
    "WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A\n"
    "PARTICULAR PURPOSE.  See the GNU General Public License for more details.\n\n"
    "You should have received a copy of the GNU General Public License along with\n"
    "this program.  If not, see <http://www.gnu.org/licenses/>.\n";

const char lossyWAV_process_description [] =
    "Process Description:\n\n"
    "lossyWAV is a near lossless audio processor which dynamically reduces the\n"
    "bitdepth of the signal on a block-by-block basis. Bitdepth reduction adds noise\n"
    "to the processed output. The amount of permissible added noise is based on\n"
    "analysis of the signal levels in the default frequency range 20Hz to 16kHz.\n\n"
    "If signals above the upper limiting frequency are at an even lower level, they\n"
    "can be swamped by the added noise. This is usually inaudible, but the behaviour\n"
    "can be changed by specifying a different --limit (in the range 10kHz to 20kHz).\n\n"
    "For many audio signals there is little content at very high frequencies and\n"
    "forcing lossyWAV to keep the added noise level lower than the content at these\n"
    "frequencies can increase the bitrate dramatically for no perceptible benefit.\n\n"
    "The noise added by the process is shaped using an adaptive method provided by\n"
    "Sebastian Gesemann. This method, as implemented in lossyWAV, aims to use the\n"
    "signal itself as the basis of the filter used for noise shaping. Adaptive noise\n"
    "shaping is enabled by default.\n";

const char lossyWAV_standard_help [] =
    "Usage   : lossyWAV <input wav file> <options>\n"
    "\nExample : lossyWAV musicfile.wav\n"
    "\nQuality Options:\n\n"
    "-q, --quality <t>    where t is one of the following (default = standard):\n"
    "    I, insane        highest quality output, suitable for transcoding;\n"
    "    E, extreme       higher quality output, suitable for transcoding;\n"
    "    H, high          high quality output, suitable for transcoding;\n"
    "    S, standard      default quality output, considered to be transparent;\n"
    "    C, economic      intermediate quality output, likely to be transparent;\n"
    "    P, portable      good quality output for DAP use, may not be transparent;\n"
    "    X, extraportable lowest quality output, not fully transparent.\n"
    "\nStandard Options:\n\n"
    "-C, --correction     write correction file for processed WAV file; default=off.\n"
    "-f, --force          forcibly over-write output file if it exists; default=off.\n"
    "-h, --help           display help.\n"
    "-L, --longhelp       display extended help.\n"
    "-M, --merge          merge existing lossy.wav and lwcdf.wav files.\n"
    "-o, --outdir <t>     destination directory for the output file(s).\n"
    "-v, --version        display the lossyWAV version number.\n"
    "-w, --writetolog     create (or add to) lossyWAV.log in the output directory.\n";

const char lossyWAV_advanced_help [] =
    "\nAdvanced Options:\n\n"
    "-                    take WAV input from STDIN.\n"
    "-c, --check          check if WAV file has already been processed; default=off.\n"
    "                     errorlevel=16 if already processed, 0 if not.\n"
    "-q, --quality <n>    quality preset (-5.0<=n<=10.0); (-5=lowest, 10=highest;\n"
    "                     default=2.5; I=10; E=7.5; H=5; S=2.5; C=0; P=-2.5; X=-5).\n"
    "--, --stdout         write WAV output to STDOUT.\n"
    "    --stdinname <t>  pseudo filename to use when input from STDIN.\n"
    "\nAdvanced Quality Options:\n\n"
    "-A, --adaptive <n/t> modify settings for Sebastian Gesemann''s adaptive noise\n"
    "                     shaping method. takes a parameter to set the order of the\n"
    "                     FIR filter, (32<=n<=96; default=64; multiple of 8 only);\n"
    "                     \"OFF\" to disable adaptive shaping; \"NOWARP\" to disable \n"
    "                     default frequency warping;\n"
    "-a, --analyses <n>   set number of FFT analysis lengths, (2<=n<=6; default=3,\n"
    "                     i.e. 32, 64 & 1024 samples. n=2, remove 32 sample FFT;\n"
    "                     n>3 add 512; n>4, add 256; n>6, add 128) nb. FFT lengths.\n"
    "                     stated are for 44.1/48kHz audio, higher sample rates will\n"
    "                     automatically increase all FFT lengths as required.\n"
    "-l, --limit <n>      set upper frequency limit to be used in analyses to n Hz;\n"
    "                     (10000<=n<=20000; default=16000).\n"
    "    --linkchannels   revert to original single bits-to-remove value for all\n"
    "                     channels rather than channel dependent bits-to-remove.\n"
    "    --maxclips <n>   set max. number of acceptable clips per channel per block;\n"
    "                     (0<=n<=16; default=3,3,3,3,3,2,2,2,2,2,1,1,1,0,0,0).\n"
    "-m, --midside        analyse 2 channel audio for mid/side content.\n"
    "    --nodccorrect    disable DC correction of audio data prior to FFT analysis,\n"
    "                     default=on; (DC offset calculated per FFT data set).\n"
    "    --scale <n>      factor to scale audio by; (0.0625<n<=8.0; default=1).\n"
    "-s, --shaping [n]    enable fixed noise shaping, takes optional parameter [n]\n"
    "                     to allow user defined shaping proportion (0.0<=n<=1.0),\n"
    "                     otherwise default to quality setting dependent value.\n"
    "                     Disables adaptive noise shaping.\n"
    "    --static <n>     set minimum-bits-to-keep-static to n bits (default=6;\n"
    "                     7<=n<=28, limited to bits-per-sample - 4).\n"
    "-U, --underlap <n>   enable underlap mode to increase number of FFT analyses\n"
    "                     performed at each FFT length, (n = 2, 4 or 8, default=2).\n"
    "\nOutput Options:\n\n"
    "    --bitdist        show distrubution of bits to remove.\n"
    "    --blockdist      show distribution of lowest / highest significant bit of\n"
    "                     input codec-blocks and bit-removed codec-blocks.\n"
    "-d, --detail         enable per block per channel bits-to-remove data display.\n"
    "-F, --freqdist       enable frequency analysis display of input data."
    "-H, --histogram      show sample value histogram (input, lossy and correction).\n"
    "    --longdist       show long frequency distribution data (input/lossy/lwcdf).\n"
    "    --perchannel     show selected distribution data per channel.\n"
    "-p, --postanalyse    enable frequency analysis display of output and\n"
    "                     correction data in addition to input data."
    "    --sampledist     show distribution of lowest / highest significant bit of\n"
    "                     input samples and bit-removed samples.\n"
    "    --spread [full]  show detailed [more detailed] results from the spreading/\n"
    "                     averaging algorithm.\n"
    "-W, --width <n>      select width of output options (79<=n<=255).\n"
    "\nSystem Options:\n\n"
    "-B, --below          set process priority to below normal.\n"
    "    --low            set process priority to low.\n"
    "-N, --nowarnings     suppress lossyWAV warnings.\n"
    "-Q, --quiet          significantly reduce screen output.\n"
    "-S, --silent         no screen output.\n";

const char lossyWAV_special_thanks [] =
    "\nSpecial thanks go to:\n\n"
    "David Robinson       for the publication of his lossyFLAC method, guidance, and\n"
    "                     the motivation to implement his method as lossyWAV.\n\n"
    "Horst Albrecht       for ABX testing, valuable support in tuning the internal\n"
    "                     presets, constructive criticism and all the feedback.\n\n"
    "Sebastian Gesemann   for the adaptive noise shaping method and the amount of\n"
    "                     help received in implementing it and also for the basis of\n"
    "                     the fixed noise shaping method.\n\n"
    "Matteo Frigo and     for libfftw3-3.dll contained in the FFTW distribution\n"
    "Steven G Johnson     (v3.2.1 or v3.2.2).\n\n"
    "Mark G Beckett       for the Delphi unit that provides an interface to the\n"
    "(Univ. of Edinburgh) relevant fftw routines in libfftw3-3.dll.\n\n"
    "Don Cross            for the Complex-FFT algorithm originally used.\n";

const int num_quality_synonyms = 6;

const double quality_synonyms_vals [ num_quality_synonyms + 1 ] = { -5.0, -2.5, 0.0, 2.5, 5.0, 7.5, 10.0 };
const char* quality_synonyms_short [ num_quality_synonyms + 1 ] = { "X", "P", "C", "S", "H", "E", "I" };
const char* quality_synonyms_long [ num_quality_synonyms + 1 ] = { "extraportable", "portable", "economic", "standard", "high", "extreme", "insane" };

std::string this_quality_synonym_long;
std::string current_parameter;
std::string parmError;
int ThisParameterNumber;
int max_open_retries = 0;

} // namespace



int ParamCount()
{
	return main_argc - 1;
}

std::string ParamStr(int index)
{
	return main_argv[index];
}

char parmchar()
{
	char result;
	result = '@';

	if ((current_parameter.length() == 2) && (current_parameter[1] == '-'))
	{
		result = current_parameter[2];
	}

	return result;
}


void parmsError()
{
	if ((ParamCount() > 0) && (parameters.help == 0))
	{
		Write(Erroutput, parmError, CRLF);
	}
	else
	{
		if (parameters.help > 1)
		{
			Write(Erroutput, std::string(lossyWAV_process_description) + CRLF);
		}

		Write(Erroutput, lossyWAV_standard_help);

		if (parameters.help > 1)
		{
			Write(Erroutput, lossyWAV_advanced_help);
		}

		Write(Erroutput, lossyWAV_special_thanks);
	}
}


bool GetNextParamStr()
{
	bool result;
	result = false;

	if (ThisParameterNumber < ParamCount())
	{
		++ ThisParameterNumber;
		current_parameter = ParamStr(ThisParameterNumber);
		result = true;
	}

	return result;
}


bool NextParameterIsParameterOrEnd()
{
	bool result;
	std::string NextParamStr;
	result = false;

	if (ThisParameterNumber == ParamCount())
	{
		result = true;
	}
	else
	{
		NextParamStr = ParamStr(ThisParameterNumber + 1);

		if (NextParamStr[1] == '-')
		{
			if (NextParamStr.length() == 1)
			{
				result = true;
			}
			else if ((int (NextParamStr[2]) > 57) || ((int (NextParamStr[2]) < 48) && (int (NextParamStr[2]) != 46)))
			{
				result = true;
			}
		}
	}

	return result;
}


void parmerror_multiple_selection()
{
	lossyWAVerror(std::string("Multiple ") + parmError + " switches given.", 0x31);
}


void parmerror_mutually_incompatible_selection()
{
	lossyWAVerror(std::string("Switches ") + parmError + " are incompatible.", 0x31);
}


void parmerror_no_value_given()
{
	lossyWAVerror(std::string("No ") + parmError + " value given.", 0x31);
}


void parmerror_val_error()
{
	lossyWAVerror(std::string("Error evaluating ") + parmError + " value given.", 0x31);
}


void check_permitted_values(int cpv_var, int cpv_low, int cpv_high)    /* overload */
{
	if ((cpv_var < cpv_low) || (cpv_var > cpv_high))
	{
		lossyWAVerror(std::string("Permitted ") + parmError + " values : " + IntToStr(cpv_low) + "<=n<=" + IntToStr(cpv_high), 0x31);
	}
}


void check_permitted_values(double cpv_var, double cpv_low, double cpv_high, int Digits)    /* overload */
{
	if ((cpv_var < cpv_low) || (cpv_var > cpv_high))
	{
		lossyWAVerror(std::string("Permitted ") + parmError + " values : " + floattostrf(cpv_low, ffFixed, 99, Digits) + "<=n<=" + floattostrf(cpv_high, ffFixed, 99, Digits), 0x31);
	}
}


bool check_parameter()
{
	bool result;
	int qs_i, valError = 0;
	result = false;

	if ((parmchar() == 'v') || (current_parameter == "--version"))
	{
		Write(Erroutput, version_string, strings.version_short, CRLF);
		throw (0);
	}

	if ((parmchar() == 'h') || (current_parameter == "--help"))
	{
		Write(Erroutput, version_string, strings.version_short, lossyWAVHead1, CRLF);
		Write(Erroutput, lossyWAV_GPL, CRLF);
		parameters.help = 1;
		parmsError();
		throw (0);
	}

	if ((parmchar() == 'L') || (current_parameter == "--longhelp"))
	{
		Write(Erroutput, version_string, strings.version_short, lossyWAVHead1, CRLF);
		Write(Erroutput, lossyWAV_GPL, CRLF);
		parameters.help = 2;
		parmsError();
		throw (0);
	}

	if ((parmchar() == 'q') || (current_parameter == "--quality"))
	{
		parmError = "quality preset";

		if (parameters.quality != -99)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		for (qs_i = 0; qs_i <= num_quality_synonyms;  ++qs_i)
		{
			this_quality_synonym_long = quality_synonyms_long[qs_i];

			if ((current_parameter == quality_synonyms_short[qs_i]) || (current_parameter == this_quality_synonym_long))
			{
				parmError = "quality preset";
				parameters.quality = quality_synonyms_vals[qs_i];
				strings.parameter = std::string("--quality ") + this_quality_synonym_long;
				result = true;
				return result;
			}
		}

		if (parameters.quality == -99)
		{
			parameters.quality = std::atof(current_parameter.c_str());
/*
			if (valError != 0)
			{
				parmerror_val_error();
			}
*/
			parameters.quality = nround10(parameters.quality, 4);
			check_permitted_values(parameters.quality, -5, 10, 4);

			if (parameters.quality == int (parameters.quality))
			{
				strings.parameter = std::string("--quality ") + IntToStr(int(parameters.quality));
			}
			else
			{
				strings.parameter = std::string("--quality ") + floattostrf(parameters.quality, ffFixed, 99, 4);
			}

			result = true;
			return result;
		}
	}

	if (current_parameter == "--maxclips")
	{
		parmError = "maximum clips";

		if (parameters.maxclips != -1)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.maxclips = std::atoi(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		check_permitted_values(parameters.maxclips, 0, 16);
		result = true;
		return result;
	}

	if ((parmchar() == '-') || (current_parameter == "--stdout"))
	{
		parmError = "output to STDOUT";

		if (parameters.STDOUTPUT == true)
		{
			parmerror_multiple_selection();
		}

		parameters.STDOUTPUT = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'B') || (current_parameter == "--below"))
	{
		parmError = "process priority";

		if (parameters.priority > 0)
		{
			parmerror_multiple_selection();
		}

		setpriority(0x00004000);
		parameters.priority = 1;
		result = true;
		return result;
	}

	if (current_parameter == "--low")
	{
		parmError = "process priority";

		if (parameters.priority > 0)
		{
			parmerror_multiple_selection();
		}

		parameters.priority = 2;
		setpriority(0x00000040);
		result = true;
		return result;
	}

	if ((parmchar() == 'd') || (current_parameter == "--detail"))
	{
		parmError = "detail";

		if (parameters.detail)
		{
			parmerror_multiple_selection();
		}

		parameters.detail = true;
		result = true;
		return result;
	}

	if (current_parameter == "--perchannel")
	{
		parmError = "per-channel data output";

		if (parameters.perchannel)
		{
			parmerror_multiple_selection();
		}

		parameters.perchannel = true;
		result = true;
		return result;
	}

	if (current_parameter == "--bitdist")
	{
		parmError = "bits-to-remove distribution";

		if (parameters.bitdist)
		{
			parmerror_multiple_selection();
		}

		parameters.bitdist = true;
		result = true;
		return result;
	}

	if (current_parameter == "--spread")
	{
		parmError = "spreading results";

		if (parameters.spread != -1)
		{
			parmerror_multiple_selection();
		}

		if (NextParameterIsParameterOrEnd())
		{
			parameters.spread = 1;
		}
		else
		{
			GetNextParamStr();

			if (!(current_parameter == "full"))
			{
				parmError = "spreading results";
				parmerror_val_error();
			}
			else
			{
				parameters.spread = 2;
			}
		}

		result = true;
		return result;
	}

	if (current_parameter == "--blockdist")
	{
		parmError = "block lsb distribution";

		if (parameters.blockdist)
		{
			parmerror_multiple_selection();
		}

		parameters.blockdist = true;
		result = true;
		return result;
	}

	if (current_parameter == "--sampledist")
	{
		parmError = "sample lsb distribution";

		if (parameters.sampledist)
		{
			parmerror_multiple_selection();
		}

		parameters.sampledist = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'F') || (current_parameter == "--freqdist"))
	{
		parmError = "input frequency distribution";

		if (parameters.freqdist)
		{
			parmerror_multiple_selection();
		}

		parameters.freqdist = true;
		result = true;
		return result;
	}

	if (current_parameter == "--longdist")
	{
		parmError = "long frequency distribution";

		if (parameters.longdist)
		{
			parmerror_multiple_selection();
		}

		parameters.longdist = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'W') || (current_parameter == "--width"))
	{
		parmError = "wide output";

		if (parameters.width != -1)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.width = std::atoi(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		check_permitted_values(parameters.width, 79, 255);
		result = true;
		return result;
	}

	if ((parmchar() == 'H') || (current_parameter == "--histogram"))
	{
		parmError = "sample value histogram";

		if (parameters.histogram)
		{
			parmerror_multiple_selection();
		}

		parameters.histogram = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'p') || (current_parameter == "--postanalyse"))
	{
		parmError = "output frequency distribution";

		if (parameters.postanalyse)
		{
			parmerror_multiple_selection();
		}

		if (! parameters.freqdist)
		{
			lossyWAVwarning("freqdist not selected before postanalyse");
			parameters.freqdist = true;
		}

		parameters.postanalyse = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'Q') || (current_parameter == "--quiet"))
	{
		parmError = "reduced output";

		if (! parameters.verbosity)
		{
			parmerror_multiple_selection();
		}

		parameters.verbosity = false;
		result = true;
		return result;
	}

	if ((parmchar() == 'S') || (current_parameter == "--silent"))
	{
		parmError = "silent running";

		if (parameters.silent == true)
		{
			parmerror_multiple_selection();
		}

		parameters.silent = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'N') || (current_parameter == "--nowarnings"))
	{
		parmError = "no warnings";

		if (! parameters.warnings)
		{
			parmerror_multiple_selection();
		}

		parameters.warnings = false;
		result = true;
		return result;
	}

	if ((parmchar() == 'c') || (current_parameter == "--check"))
	{
		parmError = "check";

		if (parameters.checking)
		{
			parmerror_multiple_selection();
		}

		parameters.checking = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'M') || (current_parameter == "--merge"))
	{
		parmError = "merge";

		if (parameters.merging)
		{
			parmerror_multiple_selection();
		}

		parameters.merging = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'w') || (current_parameter == "--writetolog"))
	{
		parmError = "write log output";

		if (parameters.writetolog)
		{
			parmerror_multiple_selection();
		}

		parameters.writetolog = true;
		result = true;
		return result;
	}

	if (current_parameter == "--stdinname")
	{
		parmError = "pseudo filename";

		if (parameters.stdinname != "")
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.stdinname = current_parameter;
		result = true;
		return result;
	}

	if ((parmchar() == 'f') || (current_parameter == "--force"))
	{
		parmError = "forced over-write";

		if (parameters.forcing)
		{
			parmerror_multiple_selection();
		}

		parameters.forcing = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'o') || (current_parameter == "--outdir"))
	{
		parmError = "destination folder";

		if (parameters.WavOutDir != "")
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.WavOutDir = current_parameter;

		if (!((parameters.WavOutDir[parameters.WavOutDir.length()] == '\\') || (parameters.WavOutDir[parameters.WavOutDir.length()] == ':')))
		{
			parameters.WavOutDir = parameters.WavOutDir + '\\';
		}

		result = true;
		return result;
	}

	if ((parmchar() == 'C') || (current_parameter == "--correction"))
	{
		parmError = "correction file";

		if (parameters.correction)
		{
			parmerror_multiple_selection();
		}

		parameters.correction = true;
		result = true;
		return result;
	}

	if ((parmchar() == 's') || (current_parameter == "--shaping"))
	{
		parmError = "adaptive and fixed noise shaping";

		if (parameters.adaptive_shaping != -2)
		{
			parmerror_mutually_incompatible_selection();
		}

		parameters.adaptive_shaping = -3;
		parameters.adaptive_shaping_warp = false;
		parmError = "fixed noise shaping";

		if (parameters.shaping != -1)
		{
			parmerror_multiple_selection();
		}

		if (! NextParameterIsParameterOrEnd())
		{
			GetNextParamStr();
			parameters.shaping = std::atof(current_parameter.c_str());
/*
			if (valError != 0)
			{
				parmerror_val_error();
			}
*/
			parameters.shaping = nround10(parameters.shaping, 4);
			check_permitted_values(parameters.shaping, 0, 1, 4);
		}
		else
		{
			parameters.shaping = -2;
		}

		result = true;
		return result;
	}

	if ((parmchar() == 'A') || (current_parameter == "--adaptive"))
	{
		parmError = "fixed and adaptive noise shaping";

		if (parameters.shaping != -1)
		{
			parmerror_mutually_incompatible_selection();
		}

		parmError = "adaptive noise shaping";

		if (NextParameterIsParameterOrEnd())
		{
			parmerror_no_value_given();
		}
		else
			do
			{
				GetNextParamStr();

				if ((current_parameter == "NOWARP") || (current_parameter == "nowarp"))
				{
					if (parameters.adaptive_shaping == -1)
					{
						parmError = "adaptive off and nowarp";
						parmerror_mutually_incompatible_selection();
					}

					if (parameters.adaptive_shaping_warp)
					{
						parameters.adaptive_shaping_warp = false;
					}
					else
					{
						parmerror_multiple_selection();
					}
				}
				else if ((current_parameter == "OFF") || (current_parameter == "off"))
				{
					if (! parameters.adaptive_shaping_warp)
					{
						parmError = "adaptive off and nowarp";
						parmerror_mutually_incompatible_selection();
					}

					if (parameters.adaptive_shaping == -2)
					{
						parameters.adaptive_shaping = -1;
					}
					else
					{
						parmerror_multiple_selection();
					}
				}
				else
				{
					if (parameters.adaptive_shaping != -2)
					{
						parmerror_multiple_selection();
					}

					parameters.adaptive_shaping = std::atoi(current_parameter.c_str());
/*
					if (valError != 0)
					{
						parmerror_val_error();
					}
*/
					check_permitted_values(parameters.adaptive_shaping, 32, 96);

					if ((parameters.adaptive_shaping & 7) != 0)
					{
						lossyWAVerror(std::string("Adaptive noise shaping value must be a multiple of 8: ") + IntToStr(parameters.adaptive_shaping) + " is invalid.", 0x31);
					}
				}
			}
			while (!(NextParameterIsParameterOrEnd()));

		result = true;
		return result;
	}

	if (current_parameter == "--static")
	{
		parmError = "minimum bits to keep (static)";

		if (parameters.Static != STATIC_MINIMUM_BITS_TO_KEEP)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.Static = std::atoi(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		check_permitted_values(parameters.Static, 7, 28);
		result = true;
		return result;
	}

	if (current_parameter == "--scale")
	{
		parmError = "scaling factor";

		if (parameters.scaling != -1)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.scaling = std::atof(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		parameters.scaling = nround10(parameters.scaling, 5);
		check_permitted_values(parameters.scaling, 0.625, 8.0, 5);
		result = true;
		return result;
	}

	if ((parmchar() == 'l') || (current_parameter == "--limit"))
	{
		parmError = "limit frequency";

		if (parameters.limit != -1)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.limit = std::atoi(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		check_permitted_values(parameters.limit, 10000, 20000);
		result = true;
		return result;
	}

	if ((parmchar() == 'a') || (current_parameter == "--analyses"))
	{
		parmError = "number of FFT analyses";

		if (parameters.analyses != 3)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.analyses = std::atoi(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		check_permitted_values(parameters.analyses, 2, 6);
		result = true;
		return result;
	}

	if ((parmchar() == 'm') || (current_parameter == "--midside"))
	{
		parmError = "midside";

		if (parameters.midside)
		{
			parmerror_multiple_selection();
		}

		parameters.midside = true;
		result = true;
		return result;
	}

	if (current_parameter == "--nodccorrect")
	{
		parmError = "DC correction";

		if (! parameters.dccorrect)
		{
			parmerror_multiple_selection();
		}

		parameters.dccorrect = false;
		result = true;
		return result;
	}

	if (current_parameter == "--linkchannels")
	{
		parmError = "link channels";

		if (parameters.linkchannels)
		{
			parmerror_multiple_selection();
		}

		parameters.linkchannels = true;
		result = true;
		return result;
	}

	if ((parmchar() == 'U') || (current_parameter == "--underlap"))
	{
		parmError = "underlap";

		if (parameters.underlap != 0)
		{
			parmerror_multiple_selection();
		}

		if (! GetNextParamStr())
		{
			parmerror_no_value_given();
		}

		parameters.underlap = std::atoi(current_parameter.c_str());
/*
		if (valError != 0)
		{
			parmerror_val_error();
		}
*/
		check_permitted_values(parameters.underlap, 2, 8);

		if ((parameters.underlap & (parameters.underlap - 1)) > 0)
		{
			lossyWAVerror("Permitted underlap values : 2, 4, 8", 0x31);
		}

		result = true;
		return result;
	}

	return result;
}


bool getParms()
{
	bool result;
	int colonpos;
	int slashpos;
	result = false;
	ThisParameterNumber = 0;
	parameters.WavInpDir = "";
	parameters.WavOutDir = "";
	parameters.wavName = "";
	parameters.stdinname = "";
	parameters.priority = 0;
	parameters.quality = -99;
	parameters.scaling = -1;
	parameters.shaping = -1;
	parameters.adaptive_shaping = -2;
	parameters.adaptive_shaping_warp = true;
	parameters.analyses = 3;
	parameters.maxclips = -1;
	parameters.help = 0;
	parameters.limit = -1;
	parameters.underlap = 0;
	parameters.Static = STATIC_MINIMUM_BITS_TO_KEEP;
	parameters.dynamic = -1;
	parameters.perchannel = false;
	parameters.bitdist = false;
	parameters.spread = -1;
	parameters.sampledist = false;
	parameters.spread = -1;
	parameters.freqdist = false;
	parameters.longdist = false;
	parameters.histogram = false;
	parameters.blockdist = false;
	parameters.checking = false;
	parameters.correction = false;
	parameters.width = -1;
	parameters.detail = false;
	parameters.forcing = false;
	parameters.linkchannels = false;
	parameters.merging = false;
	parameters.midside = false;
	parameters.dccorrect = true;
	parameters.sampledist = false;
	parameters.silent = false;
	parameters.STDINPUT = false;
	parameters.STDOUTPUT = false;
	parameters.warnings = true;
	parameters.writetolog = false;
	parameters.verbosity = true;

	if (! GetNextParamStr())
	{
		return result;
	}

	if (current_parameter == "-")
	{
		parameters.STDINPUT = true;
	}
	else if (current_parameter[1] == '-')
	{
		if ((parmchar() == 'v') || (current_parameter == "--version"))
		{
			Write(Erroutput, version_string, strings.version_short, CRLF);
			throw (0);
		}
		else
		{
			Write(Erroutput, version_string, strings.version_short, lossyWAVHead1, CRLF);
			Write(Erroutput, lossyWAV_GPL, CRLF);

			if ((parmchar() == 'h') || (current_parameter == "--help") || (parmchar() == 'L') || (current_parameter == "--longhelp"))
			{
				parameters.help = 1 + int ((current_parameter == "--longhelp") || (parmchar() == 'L'));
				parmsError();
				throw (0);
			}
			else
			{
				lossyWAVerror("No input file given.", 1);
			}
		}
	}
	else
	{
		parameters.wavName = current_parameter;
	}

	if (parameters.wavName != "")
	{
		if (parameters.wavName[1] == '"')
		{
			parameters.wavName = parameters.wavName.substr(1, parameters.wavName.length() - 1);
		}

		if (parameters.wavName[parameters.wavName.length()] == '"')
		{
			parameters.wavName = parameters.wavName.substr(0, parameters.wavName.length() - 1);
		}

		colonpos = Pos(':', parameters.wavName);

		if (colonpos > 0)
		{
			parameters.WavInpDir = parameters.wavName.substr(0, colonpos);
			parameters.wavName = parameters.wavName.substr(colonpos, parameters.wavName.length() - colonpos);
		}

		slashpos = Pos('\\', parameters.wavName);

		while (slashpos > 0)
		{
			parameters.WavInpDir = parameters.WavInpDir + parameters.wavName.substr(0, slashpos);
			parameters.wavName = parameters.wavName.substr(slashpos, parameters.wavName.length() - slashpos);
			slashpos = Pos('\\', parameters.wavName);
		}
	}

	while (GetNextParamStr())
	{
		if (! check_parameter())
		{
			parmError = std::string("%lossyWAV Error% : Incorrect option: \"") + current_parameter + "\"" + CRLF;
			return result;
		}
	}

	result = true;
	return result;
}


std::string WAVFilePrintName()
{
	std::string result;

	if (parameters.STDINPUT)
	{
		if (parameters.stdinname == "")
		{
			result = "from STDIN";
		}
		else
		{
			result = parameters.stdinname;
		}
	}
	else
	{
		result = parameters.wavName;
	}

	return result;
}


std::string wavInp()
{
	return parameters.WavInpDir + parameters.wavName;
}


std::string WAVOut()
{
	return parameters.WavOutDir + parameters.wavName;
}


std::string lossyInp()
{
	return parameters.WavInpDir + parameters.lossyName;
}


std::string lossyOut()
{
	return parameters.WavOutDir + parameters.lossyName;
}


std::string lwcdfInp()
{
	return parameters.WavInpDir + parameters.lwcdfName;
}


std::string lwcdfOut()
{
	return parameters.WavOutDir + parameters.lwcdfName;
}


void open_log_file()
{
	int open_retries;
	int lastioresult;
	open_retries = 0;

	if (! parameters.logfileopened)
	{
		filemode = 1;
		AssignFile(LogOutput, parameters.logfilename);

		if (FileExists(parameters.logfilename))
		{
			if ((filegetattr(parameters.logfilename) & 1) == 1)
			{
				lossyWAVwarning("lossyWAV.log file is read-only.");

				if (filesetattr(parameters.logfilename, filegetattr(parameters.logfilename) & 0xFE) != 0)
				{
					lossyWAVerror(std::string("Cannot gain write access to ") + parameters.logfilename, 0x31);
				}
			}

			do
			{
				Append(LogOutput);
				lastioresult = ioresult;

				if (lastioresult != 0)
				{
					++ open_retries;
					timer.DelayTime = PowersOf.Two[- 5];
					ndelay();
				}
			}
			while (!((lastioresult == 0) || (open_retries == 128)));

			if (lastioresult != 0)
			{
				lossyWAVerror("Opening lossyWAV.log", 0x31);
			}

			max_open_retries = std::max(open_retries, max_open_retries);
		}
		else
		{
			Rewrite(LogOutput);
			lastioresult = ioresult;

			if (lastioresult != 0)
			{
				lossyWAVerror("Opening lossyWAV.log", 0x31);
			}
		}
	}

	parameters.logfileopened = true;
}


void close_log_file()
{
	if (parameters.logfileopened)
	{
		flush(LogOutput);
		Close(LogOutput);
	}

	parameters.logfileopened = false;
}


void nParameter_Init()
{
	parameters.parameters_checked = false;
	parameters.logfileopened = false;

	if (! getParms())
	{
		Write(Erroutput, version_string, strings.version_short, lossyWAVHead1, CRLF);
		Write(Erroutput, lossyWAV_GPL, CRLF);

		if (parmError != "")
		{
			Write(Erroutput, parmError);
		}

		Write(Erroutput, std::string("Syntax: lossyWAV <wavfilename>") + CRLF);
		Write(Erroutput, std::string(CRLF) + "Type ''lossyWAV --help'' for more detail." + CRLF);

		throw (1);
	}

	if (parameters.quality == -99)
	{
		parameters.quality = 2.5;
		strings.parameter = "--quality standard";
	}

	if (parameters.help >= 1)
	{
		Write(Erroutput, version_string, strings.version_short, lossyWAVHead1, CRLF);
		Write(Erroutput, lossyWAV_GPL, CRLF);
		parmsError();
		throw (1);
	}

	if ((parameters.verbosity) && (! parameters.silent))
	{
		Write(Erroutput, version_string, strings.version_short, lossyWAVHead1, lossyWAVhead2);
	}

	if ((! parameters.STDINPUT) && (parameters.wavName.length() == 0))
	{
		lossyWAVerror("Name of input or output wav file missing.", 0x31);
	}

	if (parameters.STDINPUT)
	{
		parameters.wavName = "";
		parameters.lossyName = "lossywav.lossy.wav";
		parameters.lwcdfName = "lossywav.lwcdf.wav";

		if (parameters.merging)
		{
			lossyWAVerror(std::string("Merge parameter is incompatible") + CRLF + "                   with STDIN file input mode.", 0x31);
		}
	}
	else
	{
		if (! FileExists(wavInp()))
		{
			if (! FileExists(wavInp() + ".wav"))
			{
				lossyWAVerror(std::string("Input file: ") + parameters.wavName + " does not exist.", 0x31);
			}
			else
			{
				parameters.wavName = parameters.wavName + ".wav";
			}
		}
	}

	if (parameters.STDOUTPUT == true)
	{
		parameters.lossyName = "";
		parameters.lwcdfName = "";
		parameters.WavOutDir = "";

		if (parameters.correction)
		{
			lossyWAVerror(std::string("Correct parameter is incompatible") + CRLF + "                   with STDOUT file output mode.", 0x31);
		}
	}
	else
	{
		if (parameters.STDINPUT)
		{
			if (parameters.stdinname != "")
			{
				if (parameters.stdinname.length() < 5)
				{
					if (parameters.stdinname[parameters.stdinname.length() - 1] == '.') // last char
					{
						parameters.stdinname = parameters.stdinname + "wav";
					}
					else
					{
						parameters.stdinname = parameters.stdinname + ".wav";
					}
				}
				parameters.lossyName = parameters.stdinname.substr(0, parameters.stdinname.length() - 3) + "lossy." + parameters.stdinname.substr(parameters.stdinname.length() - 3, 3);
				parameters.lwcdfName = parameters.stdinname.substr(0, parameters.stdinname.length() - 3) + "lwcdf." + parameters.stdinname.substr(parameters.stdinname.length() - 3, 3);
			}
			else
			{
				parameters.lossyName = "stdin.lossy.wav";
				parameters.lwcdfName = "stdin.lwcdf.wav";
			}
		}
		else
		{
			parameters.lossyName = parameters.wavName.substr(0, parameters.wavName.length() - 3) + "lossy." + parameters.wavName.substr(parameters.wavName.length() - 3, 3);
			parameters.lwcdfName = parameters.wavName.substr(0, parameters.wavName.length() - 3) + "lwcdf." + parameters.wavName.substr(parameters.wavName.length() - 3, 3);
		}

		if ((parameters.WavOutDir != "") && (DirectoryExists(parameters.WavOutDir) == false))
		{
			lossyWAVerror(std::string("Directory : ") + parameters.WavOutDir + " cannot be accessed.", 0x31);
		}
	}

	if (! parameters.merging)
	{
		if ((parameters.STDOUTPUT == false) && (FileExists(lossyOut()) == true))
		{
			if (parameters.forcing == true)
			{
				if ((filegetattr(lossyOut()) & 1) == 1)
				{
					lossyWAVwarning("Output lossy file is read-only.");

					if (filesetattr(lossyOut(), filegetattr(lossyOut()) & 0xFE) != 0)
					{
						lossyWAVerror(std::string("Cannot gain write access to ") + lossyOut(), 0x31);
					}
				}
			}
			else
			{
				lossyWAVerror(std::string("Output file: ") + lossyOut() + " exists.", 0x31);
			}
		}

		if ((parameters.correction) && (FileExists(lwcdfOut()) == true))
		{
			if (parameters.forcing == true)
			{
				if ((filegetattr(lwcdfOut()) & 1) == 1)
				{
					lossyWAVwarning("Output correction file is read-only.");

					if (filesetattr(lwcdfOut(), filegetattr(lwcdfOut()) & 0xFE) != 0)
					{
						lossyWAVerror(std::string("Cannot gain write access to ") + lwcdfOut(), 0x31);
					}
				}
			}
			else
			{
				lossyWAVerror(std::string("Output file: ") + lwcdfOut() + " exists.", 0x31);
			}
		}
	}
	else
	{
		parameters.lossyName = parameters.wavName;

		if (parameters.lossyName.substr(parameters.lossyName.length() - 10, 10) == ".lossy.wav")
		{
			parameters.lwcdfName = parameters.wavName.substr(0, parameters.wavName.length() - 10) + ".lwcdf.wav";
			parameters.wavName = parameters.wavName.substr(0, parameters.wavName.length() - 10) + ".wav";
		}

		if (FileExists(WAVOut()) && (! parameters.forcing))
		{
			lossyWAVerror("Output merged file already exists", 0x31);
		}

		if (! FileExists(lossyInp()))
		{
			lossyWAVerror("Input lossy file not found.", 0x31);
		}

		if (! FileExists(lwcdfInp()))
		{
			lossyWAVerror("Input correction file not found.", 0x31);
		}
	}

	parameters.logfilename = parameters.WavOutDir + "lossyWAV.log";
}

// initialization: todo
static struct Init
{
	Init() {
		nParameter_Init();
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
	}
} init;

