//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <algorithm>
#include "nSGNS.h"
#include "nMaths.h"
#include "nCore.h"
#include "fftw_interface.h"
#include "nFFT_pas.h"

namespace { // anonymous

const char UnitName[] = "nSGNS";

//============================================================================
// Maximum order of FIR filter for Adaptive Noise Shaping
//============================================================================
const int MAX_FILTER_ORDER = 96;

const int tFFT_LENGTH = MAX_FFT_LENGTH_HALF + 2;
const int tFFT_Size = tFFT_LENGTH * DOUBLE_SIZE;

const float FNS_Freq[ 42 ] =
{
	0, 20, 30, 40, 50, 60, 70, 80, 90, 100,
	200, 300, 400, 500, 600, 700, 800, 900, 1000, 1500,
	2000, 2500, 3000, 3700, 4000, 5000, 6000, 7000, 8000, 9000,
	10000, 11000, 12000, 13000, 14000, 15000, 16000, 17000, 18000, 19000,
	20000, 384000
};
const float FNS_Gain[ 42 ] =
{
	12.850f, 12.783f, 12.750f, 12.717f, 12.684f, 12.651f, 12.618f, 12.585f, 12.552f, 12.519f,
	12.191f, 11.865f, 11.537f, 11.205f, 10.864f, 10.511f, 10.144f, 9.767f, 9.383f, 7.419f,
	5.397f, 2.873f, 0.854f, 0.513f, 1.642f, 5.645f, 8.692f, 10.972f, 12.739f, 14.156f,
	15.433f, 16.764f, 18.385f, 20.904f, 24.786f, 30.000f, 35.674f, 39.51f, 40.519f, 40.418f,
	40.401f, 40.401f
};

typedef double tFFT_Array_Double[ tFFT_LENGTH ];
typedef int tFFT_Array_Integer[ tFFT_LENGTH ];
typedef double tFLT_Array[ MAX_FILTER_ORDER + 2 ];

const int PACKER_SIZE = (64 - (tFFT_LENGTH* DOUBLE_SIZE * 3 + DOUBLE_SIZE + DOUBLE_SIZE + INTEGER_SIZE) & 63) / INTEGER_SIZE; // = 15 ?

struct Filter_Rec
{
	tFLT_Array k;
	tFLT_Array State;
	tFLT_Array xState;
	double e;
	double One_Over_Minus_C;
	bool Valid;
	int Packer[ PACKER_SIZE ];
};


const int Filter_SIZE = sizeof(Filter_Rec);


Filter_Rec Filter;
Filter_Rec Filter_Store[ MAX_CHANNELS + 1 ];

struct
{
	tFLT_Array A;
	tFLT_Array B;
	tFLT_Array r;
}
LD;

struct
{
	tFFT_Array_Double Shape_Curve;
	tFFT_Array_Double Shape_Total;
	tFFT_Array_Integer Negative_Freq_Index;
	tFFT_Array_Integer Length_DIV;
	tFFT_Array_Double Window;
	tFFT_Array_Double Freq;
	tFFT_Array_Double Warp;
	tFFT_Array_Double Warp_FFT_Length;
	tFFT_Array_Double Warp_Frac;
	tFFT_Array_Double Warp_1_M_Frac;
	tFFT_Array_Double Warp_Correction;
	tFFT_Array_Double Gain;
	tFFT_Array_Double FNS_Gain;
	tFFT_Array_Double Average_Ratio;
	tFFT_Array_Double Length_Factor;
	tFFT_Array_Double Length_1_M_Factor;
	tFFT_Array_Double Length_MOD_Factor;
	int Shape_Channel_Offset[ MAX_CHANNELS + 1 ];
	int Filter_Offset[ MAX_CHANNELS + 1 ];
	double Lambda;
	double OneMinusLambda2;
	double Gain_Amplitude;
	double Shaping_Factor;
	tFFT_Array_Integer Warp_Int;
	int Filter_Order;
	int Filter_Order_M1;
	void* FFTW_Plan_Inverse;
	int Skipped_Filters;
	int Limit_Freq_Bin;
	int Upper_Freq_Bin;
	bool Use_Average;
}
SGNS;

} // namespace


void Save_Filter()
{
	Filter_Store[Global.channel] = Filter;
}


void Restore_Filter()
{
	Filter = Filter_Store[Global.channel];
}


inline double ArcTan_Complex(const tDComplex& x)
{
	return std::atan2(x.Im, x.Re);
}


double warped_frequency(double Lambda, double Freq)
{
	tDComplex x, y, z;
	z.Re = std::cos(Freq);
	z.Im = std::sin(Freq);
	x.Re = z.Re + Lambda;
	x.Im = z.Im;
	y.Re = 1 + z.Re * Lambda;
	y.Im = + z.Im * Lambda;
	z.Re = (x.Re * y.Re + x.Im * y.Im);
	z.Im = (x.Im * y.Re - x.Re * y.Im);
	return ArcTan_Complex(z);
}


static void Levinson()
{
	int J, m;
	double err, km;
	LD.A[0] = 1;
	Filter.e = LD.r[0];
	m = 0;

	while ((Filter.e != 0) && (m < SGNS.Filter_Order))
	{
		++ m;
		err = LD.r[m];
		J = 1;

		while (J < m)
		{
			LD.B[m - J] = LD.A[J];
			err += LD.A[J] * LD.r[m - J];
			++ J;
		}

		km = -err / Filter.e;
		J = 1;

		while (J < m)
		{
			LD.A[J] += km * LD.B[J];
			++ J;
		}

		LD.A[m] = km;
		Filter.k[m - 1] = km;
		Filter.e *= (1.0 - km * km);
	}

	Filter.Valid = (m == SGNS.Filter_Order);
}


static void Warped_Lattice_Filter_Init()
{
	double A, o, u;
	int i;
	o = 1;
	u = 1;

	for (i = 0; i <= SGNS.Filter_Order_M1; ++i)
	{
		Filter.State[i] = 0;
		Filter.xState[i] = u;
		u = -SGNS.Lambda * u;
		A = u + Filter.k[i] * o;
		o += Filter.k[i] * u;
		u = A;
	}

	Filter.One_Over_Minus_C = -1.0 / o;
}


double Warped_Lattice_Filter_Evaluate()
{
	double A;
	double o;
	double u;
	int i;
	o = 0;
	u = 0;

	for (i = 0; i <= SGNS.Filter_Order_M1; ++i)
	{
		A = Filter.State[i];
		Filter.State[i] = u + SGNS.Lambda * A;
		u = SGNS.OneMinusLambda2 * A - SGNS.Lambda * u;
		A = o;
		o += Filter.k[i] * u;
		u += Filter.k[i] * A;
	}

	return o * Filter.One_Over_Minus_C;
}


void Warped_Lattice_Filter_Update(double fqe)
{
	int wu_i;

	for (wu_i = 0; wu_i <= SGNS.Filter_Order_M1; ++wu_i)
	{
		Filter.State[wu_i] = (fqe * Filter.xState[wu_i] + Filter.State[wu_i]);
	}
}


static void Process_Stored_Results()
{
	int ts_i;
	int ts_j;
	double ts_x;
	double ts_y;
	double hi_fact;
	double lo_fact;

	//==========================================================================
	// Fill desired shape curve.
	//==========================================================================
	ts_y = 0.0;

	for (ts_i = 0; ts_i <= SGNSFFT.FFT.length_half; ++ts_i)
	{
		ts_j = SGNS.Length_DIV[ts_i];
		hi_fact = (ts_i & 15) * OneOver[16];
		lo_fact = 1.0 - hi_fact;
		ts_x = nroot(((SGNSFFT.FFT_results_short[ts_j] * lo_fact + SGNSFFT.FFT_results_short[ts_j + 1] * hi_fact) * SGNS.Length_Factor[ts_i]) +
		              (SGNSFFT.FFT_results_long[ts_i]) * SGNS.Length_1_M_Factor[ts_i]) * SGNS.Gain[ts_i];
		SGNS.Shape_Curve[ts_i] = ts_x;
		ts_y += ts_x;
		SGNS.Shape_Total[ts_i] = ts_y;
	}

	if (SGNS.Use_Average)
	{
		//==========================================================================
		// Calculate average up to Limit_Freq_Bin
		//==========================================================================
		ts_y = SGNS.Shape_Total[SGNS.Upper_Freq_Bin] * OneOver[SGNS.Upper_Freq_Bin + 1];

		for (ts_i = SGNS.Upper_Freq_Bin + 1; ts_i <= SGNSFFT.FFT.length_half; ++ts_i)
		{
			ts_x = SGNS.Shape_Curve[ts_i];

			if (ts_x < ts_y)
			{
				SGNS.Shape_Curve[ts_i] = ts_x + (ts_y - ts_x) * SGNS.Average_Ratio[ts_i];
			}

			SGNS.Shape_Total[ts_i] = SGNS.Shape_Total[ts_i - 1] + SGNS.Shape_Curve[ts_i];
		}
	}
}


static void Fill_FFT_With_Warped_Spectrum()
{
	int ts_i;
	int ts_j;
	double ts_y;
	double ts_z;
	double ts_x;
	ts_x = 0.0;

	for (ts_i = 0; ts_i <= SGNSFFT.FFT.length_half; ++ts_i)
	{
		ts_j = SGNS.Warp_Int[ts_i];
		ts_z = SGNS.Warp_1_M_Frac[ts_i] * SGNS.Shape_Total[ts_j] + SGNS.Warp_Frac[ts_i] * SGNS.Shape_Total[ts_j + 1];
		ts_y = (ts_z - ts_x) * SGNS.Warp_Correction[ts_i];

		if ((ts_i & SGNSFFT.FFT.length_half_m1) > 0)
			/*with FFT_Array.DComplex[SGNSFFT.FFT.length-ts_i] do*/
		{
			FFT_Array.DComplex[SGNSFFT.FFT.length - ts_i].set(ts_y, 0);
		}

		/*with FFT_Array.DComplex[ts_i] do*/
		{
			FFT_Array.DComplex[ts_i].set(ts_y, 0);
		}
		ts_x = ts_z;
	}
}


static void Fill_LD_r_with_FFT_Results()
{
	int ts_i;

	for (ts_i = 0; ts_i <= SGNS.Filter_Order; ++ts_i)
	{
		LD.r[ts_i] = FFT_Array.DComplex[ts_i].Re;
	}
}


void Make_Filter()
{
	if ((parameters.adaptive_shaping != -1) && (parameters.adaptive_shaping != -3))
	{
		Process_Stored_Results();
		Fill_FFT_With_Warped_Spectrum();

		if (FFTW_Initialised())
		{
			FFTW.Execute(SGNS.FFTW_Plan_Inverse);
		}
		else
		{
			SetFFTLength(SGNSFFT.FFT.length);
			IFFT_DIT_Complex();
		}

		Fill_LD_r_with_FFT_Results();
		Levinson();

		if (Filter.Valid)
		{
			Warped_Lattice_Filter_Init();
		}
		else
		{
			Filter = Filter_Store[MAX_CHANNELS];
		}
	}
	else
	{
		Filter = Filter_Store[MAX_CHANNELS];
	}

	Save_Filter();
}


bool Filter_Valid()
{
	return Filter_Store[Global.channel].Valid;
}


int Skipped_Filters()
{
	return SGNS.Skipped_Filters;
}


void nSGNS_Initialise()
{
	int si_i, si_j, si_k;
	double this_lambda;
	double this_freq;
	double this_target_bin;
	double this_delta;
	double this_freq_limit;
	double this_average_ratio;
	double this_gain;
	double this_warp;
	double this_total;

	//============================================================================
	// Set pointer to relevant FFTW plan (if available).
	//============================================================================
	if (FFTW_Initialised())
	{
		SGNS.FFTW_Plan_Inverse = FFTW.Plan_DFT_1d(SGNSFFT.FFT.length, &FFT_Array.DReal[0], &FFT_Array.DReal[0], FFTW_BACKWARD, 0x69);
	}

	//============================================================================
	// Use high frequency average tweak?
	//============================================================================
	SGNS.Use_Average = true;

	//============================================================================
	// Set shaping factor for fixed noise shaping.
	//============================================================================
	SGNS.Shaping_Factor = parameters.shaping;

	if (SGNS.Shaping_Factor == -1)
	{
		SGNS.Shaping_Factor = 1;
	}

	//============================================================================
	// Zero number of skipped filters.
	//============================================================================
	SGNS.Skipped_Filters = 0;

	//============================================================================
	// Set amplitude of default gain curve.
	//============================================================================
	SGNS.Gain_Amplitude = 4.5;

	//============================================================================
	// Initialise filters and create channel offset tables.
	//============================================================================
	std::memset((void*) &Filter, 0, sizeof(Filter));
	std::memset((void*) &Filter_Store[0], 0, sizeof(Filter_Store));

	for (si_i = 0; si_i < MAX_CHANNELS; ++si_i)
	{
		Filter_Store[si_i].One_Over_Minus_C = 1;
		SGNS.Shape_Channel_Offset[si_i] = si_i * ((int) tFFT_Size);
		SGNS.Filter_Offset[si_i] = si_i * Filter_SIZE;
	}

	SGNS.Filter_Offset[MAX_CHANNELS] = MAX_CHANNELS * Filter_SIZE;
	Filter.One_Over_Minus_C = 1;

	//============================================================================
	// Set filter order
	//============================================================================
	if (parameters.adaptive_shaping < -1)
	{
		SGNS.Filter_Order = 64;
	}
	else
	{
		SGNS.Filter_Order = parameters.adaptive_shaping;
	}

	SGNS.Filter_Order = std::min(MAX_FILTER_ORDER, std::min(SGNS.Filter_Order, SGNSFFT.FFT.length - 16));
	SGNS.Filter_Order_M1 = std::max(0, SGNS.Filter_Order - 1);

	//============================================================================
	// Set bin range points for "Average Ratio" curve.
	//============================================================================
	SGNS.Upper_Freq_Bin = round(SGNSFFT.FFT.length * std::min(double(Global.upper_freq_limit) / Global.sample_rate, 46 * OneOver[96]));
	SGNS.Limit_Freq_Bin = round(SGNSFFT.FFT.length * std::min((HUMAN_FREQ_LIMIT + 2000.0 * Global.Codec_Block_bit_shift) / Global.sample_rate, 46 * OneOver[96]));

	//============================================================================
	// Define "average ratio" curve.
	//============================================================================
	this_average_ratio = (32 + (6 * Global.Codec_Block_bit_shift)) * OneOver[96];
	si_j = 0;
	si_k = SGNS.Upper_Freq_Bin;

	for (si_i = si_j; si_i <= si_k; ++si_i)
	{
		SGNS.Average_Ratio[si_i] = 0.0;
	}

	si_j = si_k;
	si_k = SGNS.Limit_Freq_Bin;

	for (si_i = si_j; si_i <= si_k; ++si_i)
	{
		SGNS.Average_Ratio[si_i] = nsqr(double(si_i - si_j) / (std::max(si_k - si_j, 1))) * this_average_ratio;
	}

	si_j = si_k;
	si_k = (SGNSFFT.FFT.length_half - (SGNSFFT.FFT.length_half - SGNS.Limit_Freq_Bin) / std::max(2, Global.Codec_Block_bit_shift + 2));

	for (si_i = si_j; si_i <= si_k; ++si_i)
	{
		SGNS.Average_Ratio[si_i] = this_average_ratio;
	}

	si_j = si_k;
	si_k = SGNSFFT.FFT.length_half;

	for (si_i = si_j; si_i <= si_k; ++si_i)
	{
		SGNS.Average_Ratio[si_i] = nsqr(double(si_k - si_i) / (std::max(si_k - si_j, 1))) * this_average_ratio;
	}

	//============================================================================
	// Calculate lambda value to use in frequency warping.
	//============================================================================
	if (parameters.adaptive_shaping_warp)
	{
		this_delta = SINGLE_HALF;
		this_freq_limit = std::min(double(HUMAN_FREQ_LIMIT), Global.sample_rate * 46 * OneOver[96]);
		this_lambda = SINGLE_ONE;
		this_target_bin = 46 * OneOver[96];

		do
		{
			this_freq = warped_frequency(this_lambda, this_target_bin * TwoPi) * OneOverTwoPi * Global.sample_rate;

			if (this_freq < this_freq_limit)
			{
				this_lambda -= this_delta;
			}

			if (this_freq > this_freq_limit)
			{
				this_delta *= SINGLE_HALF;
				this_lambda += this_delta;
			}
		}
		while (!((this_delta < 1e-12) || (abs(this_freq - Global.upper_freq_limit) < 1e-6) || (this_lambda > 1.0) || (this_lambda < 0.0)));
	}
	else
	{
		this_lambda = 0.0;
	}

	SGNS.Lambda = std::min(1.0, std::max(0.0, this_lambda));
	SGNS.OneMinusLambda2 = 1.0 - nsqr(SGNS.Lambda);

	//============================================================================
	// Clear stored spectral shapes
	//============================================================================
	for (si_i = 0; si_i <= SGNSFFT.FFT.length_half; ++si_i)
	{
		SGNSFFT.FFT_results_long[si_i] = 0;
		SGNSFFT.FFT_results_short[si_i] = 0;
	}

	//============================================================================
	// Calculate lambda value to use in frequency warping.
	//============================================================================
	this_warp = 0.0;

	for (si_i = 0; si_i <= SGNSFFT.FFT.length_half; ++si_i)
	{
		SGNS.Window[si_i] = Hann(si_i, SGNSFFT.FFT.bit_length);
		SGNS.Freq[si_i] = si_i * SGNSFFT.FFT.length_recip;

		SGNS.Negative_Freq_Index[si_i] = (SGNSFFT.FFT.length - si_i) * 2;

		SGNS.Warp[si_i] = warped_frequency(SGNS.Lambda, SGNS.Freq[si_i] * TwoPi) * OneOverTwoPi;

		SGNS.Warp_FFT_Length[si_i] = SGNS.Warp[si_i] * SGNSFFT.FFT.length;
		SGNS.Warp_Int[si_i] = int(SGNS.Warp_FFT_Length[si_i]);
		SGNS.Warp_Frac[si_i] = SGNS.Warp_FFT_Length[si_i] - SGNS.Warp_Int[si_i];
		SGNS.Warp_1_M_Frac[si_i] = 1 - SGNS.Warp_Frac[si_i];

		SGNS.Length_Factor[si_i] = std::pow(double(si_i) * SGNSFFT.FFT.length_half_recip, 1.5) * SINGLE_HALF;
		SGNS.Length_1_M_Factor[si_i] = SINGLE_HALF - SGNS.Length_Factor[si_i];
		SGNS.Length_Factor[si_i] = SGNS.Length_Factor[si_i] * PowersOf.Two[4];

		SGNS.Length_DIV[si_i] = si_i >> 4;
		SGNS.Length_MOD_Factor[si_i] = (si_i & 15) * OneOver[16];

		if (SGNS.Warp[si_i] - this_warp != 0)
		{
			SGNS.Warp_Correction[si_i] = 1.0 / ((SGNS.Warp[si_i] - this_warp) * SGNSFFT.FFT.length);
		}
		else
		{
			SGNS.Warp_Correction[si_i] = 1.0;
		}

		this_warp = SGNS.Warp[si_i];
	}

	//============================================================================
	// Calculate unwarped gain curve to use with adaptive shaping method.
	//============================================================================
	for (si_i = 0; si_i <= SGNSFFT.FFT.length_half; ++si_i)
	{
		this_freq = SGNS.Freq[si_i] * Global.sample_rate;

		if (this_freq <= 16000)
		{
			this_gain = 0.0;
		}
		else if (this_freq >= 20000)
		{
			this_gain = 1.0;
		}
		else // case <16000 .. 20000>:
		{
			this_gain = nsqr((this_freq - 16000) * OneOver[4000]);
		}

		SGNS.Gain[si_i] = npower2(this_gain * OneOver[10] * log2_10 * SGNS.Gain_Amplitude);
	}

	//============================================================================
	// Calculate Fixed Noise Shaping curve
	//============================================================================
	this_total = 0.0;

	for (si_i = 0; si_i <= SGNSFFT.FFT.length_half; ++si_i)
	{
		this_freq = SGNS.Freq[si_i] * Global.sample_rate;
		si_j = 0;

		while (FNS_Freq[si_j + 1] < this_freq)
		{
			++ si_j;
		}

		this_gain = (this_freq - FNS_Freq[si_j]) / (FNS_Freq[si_j + 1] - FNS_Freq[si_j]) * (FNS_Gain[si_j + 1] - FNS_Gain[si_j]) + FNS_Gain[si_j];
		SGNS.FNS_Gain[si_i] = npower2(this_gain * SGNS.Shaping_Factor * OneOver[10] * log2_10);

		this_total += SGNS.FNS_Gain[si_i];
		SGNS.Shape_Total[si_i] = this_total;
	}

	//============================================================================
	// Use the fixed shaping filter when selected (i.e. skip "make_filter") and
	// for cases where calculated adaptive shaping filter invalid.
	// (skip Process_Stored_Results stage as desired curve in SGNS.Shape_Total)
	//============================================================================
	Fill_FFT_With_Warped_Spectrum();

	if (FFTW_Initialised())
	{
		FFTW.Execute(SGNS.FFTW_Plan_Inverse);
	}
	else
	{
		SetFFTLength(SGNSFFT.FFT.length);
		IFFT_DIT_Complex();
	}

	Fill_LD_r_with_FFT_Results();
	Levinson();
	Warped_Lattice_Filter_Init();

	for (si_i = 0; si_i <= MAX_CHANNELS; ++si_i)
	{
		Global.channel = si_i;
		Save_Filter();
	}

	Global.channel = 0;
}


static struct Init
{
	Init() {
		// initialization: todo
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
		if ((FFTW_Initialised()) && (SGNS.FFTW_Plan_Inverse != NULL))
		{
			FFTW.Destroy_Plan(SGNS.FFTW_Plan_Inverse);
		}
	}
} init;
